import os
import uuid
from datetime import datetime
from collections import abc
from pathlib import Path
import numpy as np
import distutils.version
from typing import Union
import warnings

import spikeextractors as se
from spikeextractors.extraction_tools import check_get_traces_args, check_get_unit_spike_train

try:
    import pynwb
    import pandas as pd
    from pynwb import NWBHDF5IO
    from pynwb import NWBFile
    from pynwb.ecephys import ElectricalSeries, LFP
    from pynwb.ecephys import ElectrodeGroup
    from hdmf.data_utils import DataChunkIterator
    from hdmf.backends.hdf5.h5_utils import H5DataIO

    HAVE_NWB = True
except ModuleNotFoundError:
    HAVE_NWB = False

PathType = Union[str, Path, None]
ArrayType = Union[list, np.ndarray]


def check_nwb_install():
    assert HAVE_NWB, "To use the Nwb extractors, install pynwb: \n\n pip install pynwb\n\n"


def set_dynamic_table_property(dynamic_table, row_ids, property_name, values, index=False,
                               default_value=np.nan, table=False, description='no description'):
    check_nwb_install()
    if not isinstance(row_ids, list) or not all(isinstance(x, int) for x in row_ids):
        raise TypeError("'ids' must be a list of integers")
    ids = list(dynamic_table.id[:])
    if any([i not in ids for i in row_ids]):
        raise ValueError("'ids' contains values outside the range of existing ids")
    if not isinstance(property_name, str):
        raise TypeError("'property_name' must be a string")
    if len(row_ids) != len(values) and index is False:
        raise ValueError("'ids' and 'values' should be lists of same size")

    if index is False:
        if property_name in dynamic_table:
            for (row_id, value) in zip(row_ids, values):
                dynamic_table[property_name].data[ids.index(row_id)] = value
        else:
            col_data = [default_value] * len(ids)  # init with default val
            for (row_id, value) in zip(row_ids, values):
                col_data[ids.index(row_id)] = value
            dynamic_table.add_column(
                name=property_name,
                description=description,
                data=col_data,
                index=index,
                table=table
            )
    else:
        if property_name in dynamic_table:
            # TODO
            raise NotImplementedError
        else:
            dynamic_table.add_column(
                name=property_name,
                description=description,
                data=values,
                index=index,
                table=table
            )


def get_dynamic_table_property(dynamic_table, *, row_ids=None, property_name):
    all_row_ids = list(dynamic_table.id[:])
    if row_ids is None:
        row_ids = all_row_ids
    return [dynamic_table[property_name][all_row_ids.index(x)] for x in row_ids]


def get_nspikes(units_table, unit_id):
    """Return the number of spikes for chosen unit."""
    check_nwb_install()
    ids = np.array(units_table.id[:])
    indexes = np.where(ids == unit_id)[0]
    if not len(indexes):
        raise ValueError(f"{unit_id} is an invalid unit_id. Valid ids: {ids}.")
    index = indexes[0]
    if index == 0:
        return units_table['spike_times_index'].data[index]
    else:
        return units_table['spike_times_index'].data[index] - units_table['spike_times_index'].data[index - 1]


def most_relevant_ch(traces):
    """
    Calculates the most relevant channel for an Unit.
    Estimates the channel where the max-min difference of the average traces is greatest.
    traces : ndarray
        ndarray of shape (nSpikes, nChannels, nSamples)
    """
    n_channels = traces.shape[1]
    avg = np.mean(traces, axis=0)

    max_min = np.zeros(n_channels)
    for ch in range(n_channels):
        max_min[ch] = avg[ch, :].max() - avg[ch, :].min()

    relevant_ch = np.argmax(max_min)
    return relevant_ch


def update_dict(d, u):
    for k, v in u.items():
        if isinstance(v, abc.Mapping):
            d[k] = update_dict(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def list_get(l, idx, default):
    try:
        return l[idx]
    except IndexError:
        return default


def check_module(nwbfile, name: str, description: str = None):
    """
    Check if processing module exists. If not, create it. Then return module.

    Parameters
    ----------
    nwbfile: pynwb.NWBFile
    name: str
    description: str | None (optional)

    Returns
    -------
    pynwb.module
    """
    assert isinstance(nwbfile, pynwb.NWBFile), "'nwbfile' should be of type pynwb.NWBFile"
    if name in nwbfile.modules:
        return nwbfile.modules[name]
    else:
        if description is None:
            description = name
        return nwbfile.create_processing_module(name, description)


class NwbRecordingExtractor(se.RecordingExtractor):
    extractor_name = 'NwbRecording'
    has_default_locations = True
    installed = HAVE_NWB  # check at class level if installed or not
    is_writable = True
    mode = 'file'
    installation_mesg = "To use the Nwb extractors, install pynwb: \n\n pip install pynwb\n\n"

    def __init__(self, file_path, electrical_series_name='ElectricalSeries'):
        """
        Parameters
        ----------
        file_path: path to NWB file
        electrical_series_name: str, optional
        """
        assert HAVE_NWB, self.installation_mesg
        se.RecordingExtractor.__init__(self)
        self.has_unscaled = False # Needed by spikeinterface?
        self._path = str(file_path)
        import time
        with NWBHDF5IO(self._path, 'r') as io:
            nwbfile = io.read()
            if electrical_series_name is not None:
                self._electrical_series_name = electrical_series_name
            else:
                a_names = list(nwbfile.acquisition)
                if len(a_names) > 1:
                    raise ValueError('More than one acquisition found. You must specify electrical_series.')
                if len(a_names) == 0:
                    raise ValueError('No acquisitions found in the .nwb file.')
                self._electrical_series_name = a_names[0]
            es = nwbfile.acquisition[self._electrical_series_name]
            if hasattr(es, 'timestamps') and es.timestamps:
                # jfm adjusted this calc of sampling freq
                n1 = len(es.timestamps)
                n2 = min(n1, 1000)
                freq1 = (n1 - 1) / (es.timestamps[n1 - 1] - es.timestamps[0])
                freq2 = (n2 - 1) / (es.timestamps[n2 - 1] - es.timestamps[0])
                if abs(freq1 - freq2) > 100:
                    raise Exception(f'Unable to determine sampling rate. Perhaps timestamp array is not continuous. {freq1} <> {freq2}')
                self._sampling_frequency = freq1
                self.recording_start_time = es.timestamps[0]
            else:
                self._sampling_frequency = es.rate
                if hasattr(es, 'starting_time'):
                    self.recording_start_time = es.starting_time
                else:
                    self.recording_start_time = 0.

            self.num_frames = int(es.data.shape[0])
            num_channels = len(es.electrodes.table.id[:])

            # Channels gains - for RecordingExtractor, these are values to cast traces to uV
            if es.channel_conversion is not None:
                gains = es.conversion * es.channel_conversion[:] * 1e6
            else:
                gains = es.conversion * np.ones(num_channels) * 1e6

            # Extractors channel groups must be integers, but Nwb electrodes group_name can be strings
            if 'group_name' in nwbfile.electrodes.colnames:
                unique_grp_names = list(np.unique(nwbfile.electrodes['group_name'][:]))

            # Fill channel properties dictionary from electrodes table
            self.channel_ids = es.electrodes.table.id[es.electrodes.data]
            for es_ind, (channel_id, electrode_table_index) in enumerate(zip(self.channel_ids, es.electrodes.data)):
                self.set_channel_property(channel_id, 'gain', gains[es_ind])
                this_loc = []
                if 'rel_x' in nwbfile.electrodes:
                    this_loc.append(nwbfile.electrodes['rel_x'][electrode_table_index])
                    if 'rel_y' in nwbfile.electrodes:
                        this_loc.append(nwbfile.electrodes['rel_y'][electrode_table_index])
                    else:
                        this_loc.append(0)
                    self.set_channel_locations(this_loc, channel_id)

                for col in nwbfile.electrodes.colnames:
                    if isinstance(nwbfile.electrodes[col][electrode_table_index], ElectrodeGroup):
                        continue
                    elif col == 'group_name':
                        self.set_channel_groups(
                            int(unique_grp_names.index(nwbfile.electrodes[col][electrode_table_index])), channel_id)
                    elif col == 'location':
                        self.set_channel_property(channel_id, 'brain_area',
                                                  nwbfile.electrodes[col][electrode_table_index])
                    elif col in ['x', 'y', 'z', 'rel_x', 'rel_y']:
                        continue
                    else:
                        self.set_channel_property(channel_id, col, nwbfile.electrodes[col][electrode_table_index])

            # Fill epochs dictionary
            self._epochs = {}
            if nwbfile.epochs is not None:
                df_epochs = nwbfile.epochs.to_dataframe()
                self._epochs = {row['tags'][0]: {
                    'start_frame': self.time_to_frame(row['start_time']),
                    'end_frame': self.time_to_frame(row['stop_time'])}
                    for _, row in df_epochs.iterrows()}

            self._kwargs = {'file_path': str(Path(file_path).absolute()),
                            'electrical_series_name': electrical_series_name}
            self.make_nwb_metadata(nwbfile=nwbfile, es=es)

    def make_nwb_metadata(self, nwbfile, es):
        # Metadata dictionary - useful for constructing a nwb file
        self.nwb_metadata = dict()
        self.nwb_metadata['NWBFile'] = {
            'session_description': nwbfile.session_description,
            'identifier': nwbfile.identifier,
            'session_start_time': nwbfile.session_start_time,
            'institution': nwbfile.institution,
            'lab': nwbfile.lab
        }
        self.nwb_metadata['Ecephys'] = dict()
        # Update metadata with Device info
        self.nwb_metadata['Ecephys']['Device'] = []
        for dev in nwbfile.devices:
            self.nwb_metadata['Ecephys']['Device'].append({'name': dev})
        # Update metadata with ElectrodeGroup info
        self.nwb_metadata['Ecephys']['ElectrodeGroup'] = []
        for k, v in nwbfile.electrode_groups.items():
            self.nwb_metadata['Ecephys']['ElectrodeGroup'].append({
                'name': v.name,
                'description': v.description,
                'location': v.location,
                'device': v.device.name
            })
        # Update metadata with ElectricalSeries info
        self.nwb_metadata['Ecephys']['ElectricalSeries'] = []
        self.nwb_metadata['Ecephys']['ElectricalSeries'].append({
            'name': es.name,
            'description': es.description
        })

    @check_get_traces_args
    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None, return_scaled=True):
        with NWBHDF5IO(self._path, 'r') as io:
            nwbfile = io.read()
            es = nwbfile.acquisition[self._electrical_series_name]
            es_channel_ids = np.array(es.electrodes.table.id[:])[es.electrodes.data[:]].tolist()
            channel_inds = [es_channel_ids.index(id) for id in channel_ids]
            if np.array(channel_ids).size > 1 and np.any(np.diff(channel_ids) < 0):
                # get around h5py constraint that it does not allow datasets
                # to be indexed out of order
                sorted_channel_ids = np.sort(channel_ids)
                sorted_idx = np.array([list(sorted_channel_ids).index(ch) for ch in channel_ids])
                recordings = es.data[start_frame:end_frame, sorted_channel_ids].T
                traces = recordings[sorted_idx, :]
            else:
                traces = es.data[start_frame:end_frame, channel_inds].T
            # This DatasetView and lazy operations will only work within context
            # We're keeping the non-lazy version for now
            # es_view = DatasetView(es.data)  # es is an instantiated h5py dataset
            # traces = es_view.lazy_slice[start_frame:end_frame, channel_ids].lazy_transpose()
        return traces

    def get_sampling_frequency(self):
        return self._sampling_frequency

    def get_num_frames(self):
        return self.num_frames

    def get_channel_ids(self):
        return self.channel_ids.tolist()

    @staticmethod
    def add_devices(recording: se.RecordingExtractor, nwbfile=None, metadata: dict = None):
        """
        Auxiliary static method for nwbextractor.

        Adds device information to nwbfile object.
        Will always ensure nwbfile has at least one device, but multiple
        devices within the metadata list will also be created.

        Parameters
        ----------
        recording: RecordingExtractor
        nwbfile: NWBFile
            nwb file to which the recording information is to be added
        metadata: dict
            metadata info for constructing the nwb file (optional).
            Should be of the format
                metadata['Ecephys']['Device'] = [{'name': my_name,
                                                  'description': my_description}, ...]

        Missing keys in an element of metadata['Ecephys']['Device'] will be auto-populated with defaults.
        """
        if nwbfile is not None:
            assert isinstance(nwbfile, NWBFile), "'nwbfile' should be of type pynwb.NWBFile"
        defaults = dict(
            name="Device",
            description="Ecephys probe."
        )
        if metadata is None:
            metadata = dict(
                Ecephys=dict(
                    Device=[defaults]
                )
            )
        assert all([isinstance(x, dict) for x in metadata['Ecephys']['Device']]), \
            "Expected metadata['Ecephys']['Device'] to be a list of dictionaries!"

        for dev in metadata['Ecephys']['Device']:
            if dev.get('name', defaults['name']) not in nwbfile.devices:
                nwbfile.create_device(**dict(defaults, **dev))

    @staticmethod
    def add_electrode_groups(recording: se.RecordingExtractor, nwbfile=None, metadata: dict = None):
        """
        Auxiliary static method for nwbextractor.

        Adds electrode group information to nwbfile object.
        Will always ensure nwbfile has at least one electrode group.
        Will auto-generate a linked device if the specified name does not exist in the nwbfile.

        Parameters
        ----------
        recording: RecordingExtractor
        nwbfile: NWBFile
            nwb file to which the recording information is to be added
        metadata: dict
            metadata info for constructing the nwb file (optional).
            Should be of the format
                metadata['Ecephys']['ElectrodeGroup'] = [{'name': my_name,
                                                          'description': my_description,
                                                          'location': electrode_location,
                                                          'device_name': my_device_name}, ...]

        Missing keys in an element of metadata['Ecephys']['ElectrodeGroup'] will be auto-populated with defaults.

        Group names set by RecordingExtractor channel properties will also be included with passed metadata,
        but will only use default description and location.
        """
        if nwbfile is not None:
            assert isinstance(nwbfile, NWBFile), "'nwbfile' should be of type pynwb.NWBFile"
        if len(nwbfile.devices) == 0:
            se.NwbRecordingExtractor.add_devices(recording, nwbfile)
        defaults = dict(
            name="Electrode Group",
            description="no description",
            location="unknown",
            device_name="Device"
        )
        if metadata is None:
            metadata = dict(
                Ecephys=dict(
                    ElectrodeGroup=[]
                )
            )
        assert all([isinstance(x, dict) for x in metadata['Ecephys']['ElectrodeGroup']]), \
            "Expected metadata['Ecephys']['ElectrodeGroup'] to be a list of dictionaries!"

        for grp in metadata['Ecephys']['ElectrodeGroup']:
            device_name = grp.get('device_name', defaults['device_name'])
            if grp.get('name', defaults['name']) not in nwbfile.electrode_groups:
                if device_name not in nwbfile.devices:
                    new_device = dict(
                        Ecephys=dict(
                            Device=dict(
                                name=device_name
                            )
                        )
                    )
                    se.NwbRecordingExtractor.add_devices(recording, nwbfile, metadata=new_device)
                    warnings.warn(f"Device \'{device_name}\' not detected in "
                                  "attempted link to electrode group! Automatically generating.")
                electrode_group_kwargs = dict(defaults, **grp)
                electrode_group_kwargs.pop('device_name')
                electrode_group_kwargs.update(device=nwbfile.devices[device_name])
                nwbfile.create_electrode_group(**electrode_group_kwargs)

        if not nwbfile.electrode_groups:
            device_name = list(nwbfile.devices.keys())[0]
            device = nwbfile.devices[device_name]
            if len(nwbfile.devices) > 1:
                warnings.warn("More than one device found when adding electrode group "
                              f"via channel properties: using device \'{device_name}\'. To use a "
                              "different device, indicate it the metadata argument.")

            electrode_group_kwargs = dict(defaults)
            electrode_group_kwargs.pop('device_name')
            electrode_group_kwargs.update(device=device)
            for grp_name in np.unique(recording.get_channel_groups()).tolist():
                electrode_group_kwargs.update(name=str(grp_name))
                nwbfile.create_electrode_group(**electrode_group_kwargs)

    @staticmethod
    def add_electrodes(recording: se.RecordingExtractor, nwbfile=None, metadata: dict = None):
        """
        Auxiliary static method for nwbextractor.

        Adds channels from recording object as electrodes to nwbfile object.

        Parameters
        ----------
        recording: RecordingExtractor
        nwbfile: NWBFile
            nwb file to which the recording information is to be added
        metadata: dict
            metadata info for constructing the nwb file (optional).
            Should be of the format
                metadata['Ecephys']['Electrodes'] = [{'name': my_name,
                                                      'description': my_description,
                                                      'data': [my_electrode_data]}, ...]
            where [my_electrode_data] is a list in one-to-one correspondence with
            the nwbfile electrode ids and RecordingExtractor channel ids.

        Missing keys in an element of metadata['Ecephys']['ElectrodeGroup'] will be auto-populated with defaults
        whenever possible.

        If 'my_name' is set to one of the required fields for nwbfile
        electrodes (id, x, y, z, imp, loccation, filtering, group_name),
        then the metadata will override their default values.

        Setting 'my_name' to metadata field 'group' is not supported as the linking to
        nwbfile.electrode_groups is handled automatically; please specify the string 'group_name' in this case.

        If no group information is passed via metadata, automatic linking to existing electrode groups,
        possibly including the default, will occur.
        """
        if nwbfile is not None:
            assert isinstance(nwbfile, NWBFile), "'nwbfile' should be of type pynwb.NWBFile"
        if nwbfile.electrode_groups is None:
            se.NwbRecordingExtractor.add_electrode_groups(recording, nwbfile, metadata)
        # For older versions of pynwb, we need to manually add these columns
        if distutils.version.LooseVersion(pynwb.__version__) < '1.3.0':
            if nwbfile.electrodes is None or 'rel_x' not in nwbfile.electrodes.colnames:
                nwbfile.add_electrode_column('rel_x', 'x position of electrode in electrode group')
            if nwbfile.electrodes is None or 'rel_y' not in nwbfile.electrodes.colnames:
                nwbfile.add_electrode_column('rel_y', 'y position of electrode in electrode group')
        defaults = dict(
            id=np.nan,
            x=np.nan,
            y=np.nan,
            z=np.nan,
            # There doesn't seem to be a canonical default for impedence, if missing.
            # The NwbRecordingExtractor follows the -1.0 convention, other scripts sometimes use np.nan
            imp=-1.0,
            location="unknown",
            filtering="none",
            group_name="Electrode Group"
        )
        if metadata is None:
            metadata = dict(
                Ecephys=dict(
                    Electrodes=[]
                )
            )
        assert all([isinstance(x, dict) and set(x.keys()) == set(['name', 'description', 'data'])
                    and isinstance(x['data'], list) for x in metadata['Ecephys']['Electrodes']]), \
            "Expected metadata['Ecephys']['Electrodes'] to be a list of dictionaries!"
        assert all([x['name'] != 'group' for x in metadata['Ecephys']['Electrodes']]), \
            "Passing metadata field 'group' is depricated; pass group_name instead!"

        if nwbfile.electrodes is None:
            nwb_elec_ids = []
        else:
            nwb_elec_ids = nwbfile.electrodes.id.data[:]
        for metadata_column in metadata['Ecephys']['Electrodes']:
            if (nwbfile.electrodes is None or metadata_column['name'] not in nwbfile.electrodes.colnames) \
                    and metadata_column['name'] != 'group_name':
                nwbfile.add_electrode_column(
                    name=str(metadata_column['name']),
                    description=str(metadata_column['description'])
                )

        for j, channel_id in enumerate(recording.get_channel_ids()):
            if channel_id not in nwb_elec_ids:
                electrode_kwargs = dict(defaults)
                electrode_kwargs.update(id=channel_id)

                # recording.get_channel_locations defaults to np.nan if there are none
                location = recording.get_channel_locations(channel_ids=channel_id)[0]
                if all([not np.isnan(loc) for loc in location]):
                    # property 'location' of RX channels corresponds to rel_x and rel_ y of NWB electrodes
                    electrode_kwargs.update(
                        dict(
                            rel_x=float(location[0]),
                            rel_y=float(location[1])
                        )
                    )

                for metadata_column in metadata['Ecephys']['Electrodes']:
                    if metadata_column['name'] == 'group_name':
                        group_name = list_get(metadata_column['data'], j, defaults['group_name'])
                        if group_name not in nwbfile.electrode_groups:
                            warnings.warn(f"Electrode group for electrode {channel_id} was not "
                                          "found in the nwbfile! Automatically adding.")
                            missing_group_metadata = dict(
                                Ecephys=dict(
                                    ElectrodeGroup=dict(
                                        name=group_name,
                                        description="no description",
                                        location="unknown",
                                        device_name="Device"
                                    )
                                )
                            )
                            se.NwbRecordingExtractor.add_electrode_groups(recording, nwbfile, missing_group_metadata)
                        electrode_kwargs.update(
                            dict(
                                group=nwbfile.electrode_groups[group_name],
                                group_name=group_name
                            )
                        )
                    else:
                        if metadata_column['name'] in defaults:
                            electrode_kwargs.update({
                                metadata_column['name']: list_get(metadata_column['data'], j,
                                                                  defaults[metadata_column['name']])
                            })
                        else:
                            if j < len(metadata_column['data']):
                                electrode_kwargs.update({
                                    metadata_column['name']: metadata_column['data'][j]
                                })
                            else:
                                metadata_column_name = metadata_column['name']
                                warnings.warn(f"Custom column {metadata_column_name} "
                                              f"has incomplete data for channel id [{j}] and no "
                                              "set default! Electrode will not be added.")
                                continue

                if not any([x.get('name', '') == 'group_name' for x in metadata['Ecephys']['Electrodes']]):
                    group_id = recording.get_channel_groups(channel_ids=channel_id)[0]
                    if group_id in range(len(nwbfile.electrode_groups)):
                        group_name = list(nwbfile.electrode_groups.keys())[group_id]
                        electrode_kwargs.update(
                            dict(
                                group=nwbfile.electrode_groups[group_name],
                                group_name=group_name
                            )
                        )
                    else:
                        warnings.warn("No metadata was passed specifying the electrode group for "
                                      f"electrode {channel_id}, and the internal recording channel group was "
                                      f"assigned a value ({group_id}) outside the indices of the electrode "
                                      "groups in the nwbfile! Electrode will not be added.")
                        continue

                nwbfile.add_electrode(**electrode_kwargs)

        # property 'gain' should not be in the NWB electrodes_table
        # property 'brain_area' of RX channels corresponds to 'location' of NWB electrodes
        channel_prop_names = set(recording.get_shared_channel_property_names()) - set(nwbfile.electrodes.colnames) \
            - set(['gain', 'location'])
        for channel_prop_name in channel_prop_names:
            for channel_id in recording.get_channel_ids():
                val = recording.get_channel_property(channel_id, channel_prop_name)
                descr = 'no description'
                if channel_prop_name == 'brain_area':
                    channel_prop_name = 'location'
                    descr = 'brain area location'
                set_dynamic_table_property(
                    dynamic_table=nwbfile.electrodes,
                    row_ids=[channel_id],
                    property_name=channel_prop_name,
                    values=[val],
                    default_value=np.nan,
                    description=descr
                )

    @staticmethod
    def add_electrical_series(recording: se.RecordingExtractor, nwbfile=None, metadata: dict = None,
                              buffer_mb: int = 500, use_timestamps: bool = False, write_as_lfp: bool = False):
        """
        Auxiliary static method for nwbextractor.

        Adds traces from recording object as ElectricalSeries to nwbfile object.

        Parameters
        ----------
        recording: RecordingExtractor
        nwbfile: NWBFile
            nwb file to which the recording information is to be added
        metadata: dict
            metadata info for constructing the nwb file (optional).
            Should be of the format
                metadata['Ecephys']['ElectricalSeries'] = {'name': my_name,
                                                           'description': my_description}
        buffer_mb: int (optional, defaults to 500MB)
            maximum amount of memory (in MB) to use per iteration of the
            DataChunkIterator (requires traces to be memmap objects)
        use_timestamps: bool (optional, defaults to False)
            If True, the timestamps are saved to the nwb file using recording.frame_to_time(). If False (defualut),
            the sampling rate is used.
        write_as_lfp: bool (optional, defaults to False)
            If True, writes the traces under a processing LFP module in the NWBFile instead of acquisition.

        Missing keys in an element of metadata['Ecephys']['ElectrodeGroup'] will be auto-populated with defaults
        whenever possible.
        """
        if nwbfile is not None:
            assert isinstance(nwbfile, NWBFile), "'nwbfile' should be of type pynwb.NWBFile!"
        assert buffer_mb > 10, "'buffer_mb' should be at least 10MB to ensure data can be chunked!"
        if not nwbfile.electrodes:
            se.NwbRecordingExtractor.add_electrodes(recording, nwbfile, metadata)

        raw_defaults = dict(
            name="ElectricalSeries",
            description="Raw acquired data."
        )
        lfp_defaults = dict(
            name="LFP",
            description="Local field potential signal."
        )
        if metadata is None:
            metadata = dict(Ecephys=dict())
            if not write_as_lfp:
                metadata['Ecephys'].update(
                    ElectricalSeries=dict(raw_defaults)
                )
            else:
                metadata['Ecephys'].update(
                    LFPElectricalSeries=dict(lfp_defaults)
                )
        assert ('ElectricalSeries' in metadata['Ecephys']
                and isinstance(metadata['Ecephys']['ElectricalSeries'], dict)) \
            or ('LFPElectricalSeries' in metadata['Ecephys']
                and isinstance(metadata['Ecephys']['LFPElectricalSeries'], dict)), \
            "Expected either metadata['Ecephys']['ElectricalSeries'] or metadata['Ecephys']['LFPElectricalSeries'] " \
            "to be a dictionary!"

        if not write_as_lfp:
            es_name = metadata['Ecephys']['ElectricalSeries'].get('name', raw_defaults['name'])
            assert es_name not in nwbfile.acquisition, \
                f"Raw ElectricalSeries '{es_name}' is already written in the NWBFile!"
        else:
            es_name = metadata['Ecephys']['LFPElectricalSeries'].get('name', lfp_defaults['name'])
            ecephys_mod = check_module(
                nwbfile,
                'ecephys',
                "Intermediate data from extracellular electrophysiology recordings, e.g., LFP."
            )
            if 'LFP' not in ecephys_mod.data_interfaces:
                ecephys_mod.add_data_interface(LFP(name='LFP'))
        channel_ids = recording.get_channel_ids()

        table_ids = [list(nwbfile.electrodes.id[:]).index(id) for id in channel_ids]
        electrode_table_region = nwbfile.create_electrode_table_region(
            region=table_ids,
            description="electrode_table_region"
        )

        # channels gains - for RecordingExtractor, these are values to cast traces to uV
        # for nwb, the conversions (gains) cast the data to Volts
        # To get traces in Volts = data*channel_conversion*conversion
        gains = np.squeeze([recording.get_channel_gains(channel_ids=[ch])
                            if 'gain' in recording.get_channel_property_names(channel_id=ch) else 1
                            for ch in channel_ids])
        if len(np.unique(gains)) == 1:  # if all gains are equal
            scalar_conversion = np.unique(gains)[0] * 1e-6
            channel_conversion = None
        else:
            scalar_conversion = 1.
            channel_conversion = gains * 1e-6

        if isinstance(recording.get_traces(), np.memmap):
            n_bytes = np.dtype(recording.get_dtype()).itemsize
            buffer_size = int(buffer_mb * 1e6) // (recording.get_num_channels() * n_bytes)
            ephys_data = DataChunkIterator(
                data=recording.get_traces().T,  # nwb standard is time as zero axis
                buffer_size=buffer_size
            )
        else:
            def data_generator(recording, channels_ids):
                # generates data chunks for iterator
                for id in channels_ids:
                    data = recording.get_traces(channel_ids=[id]).flatten()
                    yield data
            ephys_data = DataChunkIterator(
                data=data_generator(
                    recording=recording,
                    channels_ids=channel_ids
                ),
                iter_axis=1,  # nwb standard is time as zero axis
                maxshape=(recording.get_num_frames(), recording.get_num_channels())
            )

        eseries_kwargs = dict(
            name=es_name,
            data=H5DataIO(ephys_data, compression="gzip"),
            electrodes=electrode_table_region,
            conversion=scalar_conversion,
            channel_conversion=channel_conversion,
            comments="Generated from SpikeInterface::NwbRecordingExtractor"
        )
        if not use_timestamps:
            eseries_kwargs.update(
                starting_time=recording.frame_to_time(0),
                rate=recording.get_sampling_frequency()
            )
        else:
            eseries_kwargs.update(
                timestamps=H5DataIO(
                    recording.frame_to_time(np.arange(recording.get_num_frames())),
                    compression="gzip"
                )
            )

        if not write_as_lfp:
            eseries_kwargs.update(
                description=metadata['Ecephys']['ElectricalSeries'].get('description', raw_defaults['description'])
            )
            nwbfile.add_acquisition(ElectricalSeries(**eseries_kwargs))
        else:
            eseries_kwargs.update(
                description=metadata['Ecephys']['LFPElectricalSeries'].get('description', lfp_defaults['description'])
            )
            ecephys_mod.data_interfaces['LFP'].add_electrical_series(ElectricalSeries(**eseries_kwargs))

    @staticmethod
    def add_epochs(recording: se.RecordingExtractor, nwbfile=None,
                   metadata: dict = None):
        '''
        Auxiliary static method for nwbextractor.
        Adds epochs from recording object to nwbfile object.

        Parameters
        ----------
        recording: RecordingExtractor
        nwbfile: NWBFile
            nwb file to which the recording information is to be added
        metadata: dict
            metadata info for constructing the nwb file (optional).
        '''
        if nwbfile is not None:
            assert isinstance(nwbfile, NWBFile), "'nwbfile' should be of type pynwb.NWBFile"

        # add/update epochs
        for epoch_name in recording.get_epoch_names():
            epoch = recording.get_epoch_info(epoch_name)
            if nwbfile.epochs is None:
                nwbfile.add_epoch(
                    start_time=recording.frame_to_time(epoch['start_frame']),
                    stop_time=recording.frame_to_time(epoch['end_frame']),
                    tags=epoch_name
                )
            else:
                if [epoch_name] in nwbfile.epochs['tags'][:]:
                    ind = nwbfile.epochs['tags'][:].index([epoch_name])
                    nwbfile.epochs['start_time'].data[ind] = recording.frame_to_time(epoch['start_frame'])
                    nwbfile.epochs['stop_time'].data[ind] = recording.frame_to_time(epoch['end_frame'])
                else:
                    nwbfile.add_epoch(
                        start_time=recording.frame_to_time(epoch['start_frame']),
                        stop_time=recording.frame_to_time(epoch['end_frame']),
                        tags=epoch_name
                    )

    @staticmethod
    def add_all_to_nwbfile(recording: se.RecordingExtractor, nwbfile=None,
                           use_timestamps: bool = False, metadata: dict = None, write_as_lfp: bool = False):
        '''
        Auxiliary static method for nwbextractor.
        Adds all recording related information from recording object and metadata
        to the nwbfile object.

        Parameters
        ----------
        recording: RecordingExtractor
        nwbfile: NWBFile
            nwb file to which the recording information is to be added
        use_timestamps: bool
            If True, the timestamps are saved to the nwb file using recording.frame_to_time(). If False (defualut),
            the sampling rate is used.
        metadata: dict
            metadata info for constructing the nwb file (optional).
            Check the auxiliary function docstrings for more information
            about metadata format.
        write_as_lfp: bool (optional, defaults to False)
            If True, writes the traces under a processing LFP module in the NWBFile instead of acquisition.
        '''
        if nwbfile is not None:
            assert isinstance(nwbfile, NWBFile), "'nwbfile' should be of type pynwb.NWBFile"

        se.NwbRecordingExtractor.add_devices(
            recording=recording,
            nwbfile=nwbfile,
            metadata=metadata
        )

        se.NwbRecordingExtractor.add_electrode_groups(
            recording=recording,
            nwbfile=nwbfile,
            metadata=metadata
        )

        # Add electrodes
        se.NwbRecordingExtractor.add_electrodes(
            recording=recording,
            nwbfile=nwbfile,
            metadata=metadata
        )

        # Add electrical series
        se.NwbRecordingExtractor.add_electrical_series(
            recording=recording,
            nwbfile=nwbfile,
            use_timestamps=use_timestamps,
            metadata=metadata,
            write_as_lfp=write_as_lfp
        )

        # Add epochs
        se.NwbRecordingExtractor.add_epochs(
            recording=recording,
            nwbfile=nwbfile,
            metadata=metadata
        )

    @staticmethod
    def write_recording(recording: se.RecordingExtractor, save_path: PathType = None,
                        nwbfile=None, use_timestamps: bool = False, metadata: dict = None, write_as_lfp: bool = False):
        '''
        Writes all recording related information from recording object and metadata
        to either a saved nwbfile (with save_path specified) or directly to an
        nwbfile object (if nwbfile specified).

        Parameters
        ----------
        recording: RecordingExtractor
        save_path: PathType
            Required if an nwbfile is not passed. Must be the path to the nwbfile
            being appended, otherwise one is created and written.
        nwbfile: NWBFile
            Required if a save_path is not specified. If passed, this function
            will fill the relevant fields within the nwbfile. E.g., calling
            spikeextractors.NwbRecordingExtractor.write_recording(
                my_recording_extractor, my_nwbfile
            )
            will result in the appropriate changes to the my_nwbfile object.
        use_timestamps: bool
            If True, the timestamps are saved to the nwb file using recording.frame_to_time(). If False (defualut),
            the sampling rate is used.
        metadata: dict
            metadata info for constructing the nwb file (optional). Should be
            of the format
                metadata['Ecephys'] = {}
            with keys of the forms
                metadata['Ecephys']['Device'] = [{'name': my_name,
                                                  'description': my_description}, ...]
                metadata['Ecephys']['ElectrodeGroup'] = [{'name': my_name,
                                                          'description': my_description,
                                                          'location': electrode_location,
                                                          'device_name': my_device_name}, ...]
                metadata['Ecephys']['Electrodes'] = [{'name': my_name,
                                                      'description': my_description,
                                                      'data': [my_electrode_data]}, ...]
                metadata['Ecephys']['ElectricalSeries'] = {'name': my_name,
                                                           'description': my_description}
        write_as_lfp: bool (optional, defaults to False)
            If True, writes the traces under a processing LFP module in the NWBFile instead of acquisition.
        '''
        assert HAVE_NWB, NwbRecordingExtractor.installation_mesg

        if nwbfile is not None:
            assert isinstance(nwbfile, NWBFile), "'nwbfile' should be of type pynwb.NWBFile"

        assert distutils.version.LooseVersion(pynwb.__version__) >= '1.3.3', \
            "'write_recording' not supported for version < 1.3.3. Run pip install --upgrade pynwb"

        assert save_path is None or nwbfile is None, \
            "Either pass a save_path location, or nwbfile object, but not both!"

        # Update any previous metadata with user passed dictionary
        if hasattr(recording, 'nwb_metadata'):
            metadata = update_dict(recording.nwb_metadata, metadata)

        if nwbfile is None:
            if Path(save_path).is_file():
                read_mode = 'r+'
            else:
                read_mode = 'w'

            with NWBHDF5IO(save_path, mode=read_mode) as io:
                if read_mode == 'r+':
                    nwbfile = io.read()
                else:
                    # Default arguments will be over-written if contained in metadata
                    nwbfile_kwargs = dict(session_description='no description',
                                          identifier=str(uuid.uuid4()),
                                          session_start_time=datetime.now())
                    if metadata is not None and 'NWBFile' in metadata:
                        nwbfile_kwargs.update(metadata['NWBFile'])
                    nwbfile = NWBFile(**nwbfile_kwargs)

                    se.NwbRecordingExtractor.add_all_to_nwbfile(
                        recording=recording,
                        nwbfile=nwbfile,
                        metadata=metadata,
                        use_timestamps=use_timestamps,
                        write_as_lfp=write_as_lfp
                    )

                # Write to file
                io.write(nwbfile)
        else:
            se.NwbRecordingExtractor.add_all_to_nwbfile(
                recording=recording,
                nwbfile=nwbfile,
                use_timestamps=use_timestamps,
                metadata=metadata,
                write_as_lfp=write_as_lfp
            )


class NwbSortingExtractor(se.SortingExtractor):
    extractor_name = 'NwbSortingExtractor'
    installed = HAVE_NWB  # check at class level if installed or not
    is_writable = True
    mode = 'file'
    installation_mesg = "To use the Nwb extractors, install pynwb: \n\n pip install pynwb\n\n"

    def __init__(self, file_path, electrical_series=None, sampling_frequency=None):
        """
        Parameters
        ----------
        path: path to NWB file
        electrical_series: pynwb.ecephys.ElectricalSeries object
        """
        assert HAVE_NWB, self.installation_mesg
        se.SortingExtractor.__init__(self)
        self._path = str(file_path)
        with NWBHDF5IO(self._path, 'r') as io:
            nwbfile = io.read()
            if sampling_frequency is None:
                # defines the electrical series from where the sorting came from
                # important to know the sampling_frequency
                if electrical_series is None:
                    if len(nwbfile.acquisition) > 1:
                        raise Exception('More than one acquisition found. You must specify electrical_series.')
                    if len(nwbfile.acquisition) == 0:
                        raise Exception('No acquisitions found in the .nwb file.')
                    es = list(nwbfile.acquisition.values())[0]
                else:
                    es = electrical_series

                if hasattr(es, 'timestamps') and es.timestamps:
                    # jfm adjusted this calc of sampling freq
                    n1 = len(es.timestamps)
                    n2 = min(n1, 1000)
                    freq1 = (n1 - 1) / (es.timestamps[n1 - 1] - es.timestamps[0])
                    freq2 = (n2 - 1) / (es.timestamps[n2 - 1] - es.timestamps[0])
                    if abs(freq1 - freq2) > 100:
                        raise Exception(f'Unable to determine sampling rate. Perhaps timestamp array is not continuous. {freq1} <> {freq2}')
                    self._sampling_frequency = freq1
                    self.recording_start_time = es.timestamps[0]
                else:
                    self._sampling_frequency = es.rate
                    if hasattr(es, 'starting_time'):
                        self.recording_start_time = es.starting_time
                    else:
                        self.recording_start_time = 0.
            else:
                self._sampling_frequency = sampling_frequency

            # get all units ids
            units_ids = nwbfile.units.id[:]

            # store units properties and spike features to dictionaries
            all_pr_ft = list(nwbfile.units.colnames)
            all_names = [i.name for i in nwbfile.units.columns]
            for item in all_pr_ft:
                if item == 'spike_times':
                    continue
                # test if item is a unit_property or a spike_feature
                if item + '_index' in all_names:  # if it has index, it is a spike_feature
                    for u_id in units_ids:
                        ind = list(units_ids).index(u_id)
                        self.set_unit_spike_features(u_id, item, nwbfile.units[item][ind])
                else:  # if it is unit_property
                    for u_id in units_ids:
                        ind = list(units_ids).index(u_id)
                        if isinstance(nwbfile.units[item][ind], pd.DataFrame):
                            prop_value = nwbfile.units[item][ind].index[0]
                        else:
                            prop_value = nwbfile.units[item][ind]

                        if isinstance(prop_value, (list, np.ndarray)):
                            self.set_unit_property(u_id, item, prop_value)
                        else:
                            if prop_value == prop_value:  # not nan
                                self.set_unit_property(u_id, item, prop_value)

            # Fill epochs dictionary
            self._epochs = {}
            if nwbfile.epochs is not None:
                df_epochs = nwbfile.epochs.to_dataframe()
                self._epochs = {row['tags'][0]: {
                    'start_frame': self.time_to_frame(row['start_time']),
                    'end_frame': self.time_to_frame(row['stop_time'])}
                    for _, row in df_epochs.iterrows()}
        self._kwargs = {'file_path': str(Path(file_path).absolute()), 'electrical_series': electrical_series,
                        'sampling_frequency': sampling_frequency}

    def get_unit_ids(self):
        '''This function returns a list of ids (ints) for each unit in the sorted result.
        Returns
        ----------
        unit_ids: array_like
            A list of the unit ids in the sorted result (ints).
        '''
        check_nwb_install()
        with NWBHDF5IO(self._path, 'r') as io:
            nwbfile = io.read()
            unit_ids = [int(i) for i in nwbfile.units.id[:]]
        return unit_ids

    @check_get_unit_spike_train
    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        start_frame, end_frame = self._cast_start_end_frame(start_frame, end_frame)
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = np.Inf
        check_nwb_install()
        with NWBHDF5IO(self._path, 'r') as io:
            nwbfile = io.read()
            # chosen unit and interval
            times = nwbfile.units['spike_times'][list(nwbfile.units.id[:]).index(unit_id)][:]
            # spike times are measured in samples
            frames = self.time_to_frame(times)
        return frames[(frames > start_frame) & (frames < end_frame)]

    def time_to_frame(self, time):
        return np.round(time * self.get_sampling_frequency()).astype('int64')

    def frame_to_time(self, frame):
        return np.round(frame / self.get_sampling_frequency(), 6)

    @staticmethod
    def write_units(sorting: se.SortingExtractor, nwbfile,
                    property_descriptions: dict, timestamps: ArrayType = None):
        '''
        Helper function for write_sorting.
        '''
        unit_ids = sorting.get_unit_ids()
        fs = sorting.get_sampling_frequency()

        all_properties = set()
        all_features = set()
        for unit_id in unit_ids:
            all_properties.update(sorting.get_unit_property_names(unit_id))
            all_features.update(sorting.get_unit_spike_feature_names(unit_id))

        if property_descriptions is not None:
            for pr in all_properties:
                if pr not in property_descriptions:
                    print(f"Warning: description for property {pr} not found in property_descriptions. "
                          f"Setting description to 'no description'")
        else:
            property_descriptions = {}

        # If no Units present in nwb file
        if nwbfile.units is None:
            # check that array properties have the same shape across units
            skip_properties = []
            property_shapes = {}
            for pr in all_properties:
                shapes = []
                for unit_id in unit_ids:
                    if pr in sorting.get_unit_property_names(unit_id):
                        prop_value = sorting.get_unit_property(unit_id, pr)
                        if isinstance(prop_value, (int, np.integer, float, np.float, str, bool)):
                            shapes.append(1)
                        elif isinstance(prop_value, (list, np.ndarray)):
                            if np.array(prop_value).ndim == 1:
                                shapes.append(len(prop_value))
                            else:
                                shapes.append(np.array(prop_value).shape)
                        elif isinstance(prop_value, dict):
                            print(f"Skipping property '{pr}' because dictionaries are not supported.")
                            skip_properties.append(pr)
                            break
                    else:
                        shapes.append(np.nan)
                property_shapes[pr] = shapes

            for pr in property_shapes.keys():
                if not np.all([elem == property_shapes[pr][0] for elem in property_shapes[pr]]):
                    print(f"Skipping property '{pr}' because it has variable size across units.")
                    skip_properties.append(pr)

            for pr in all_properties:
                # Special case of setting max_electrodes requires a table to be
                # passed to become a dynamic table region
                if pr not in skip_properties:
                    if pr in ['max_channel', 'max_electrode']:
                        if nwbfile.electrodes is None:
                            print('Warning: Attempted to make a custom column for max_channel '
                                  'or max_electrode, but there are no electrodes to reference! '
                                  'Column will not be added.')
                        else:
                            nwbfile.add_unit_column(
                                name=pr,
                                description=property_descriptions.get(pr, 'no description'),
                                table=nwbfile.electrodes
                            )
                    else:
                        nwbfile.add_unit_column(pr, property_descriptions.get(pr, 'no description'))

            for unit_id in unit_ids:
                unit_kwargs = {}
                # spike trains withinin the SortingExtractor object are not scaled by sampling frequency
                if timestamps is not None:
                    spike_train_frames = sorting.get_unit_spike_train(unit_id=unit_id)
                    assert spike_train_frames[-1] < len(timestamps), "Number of 'timestamps' differs from number of " \
                                                                     "'frames'"
                    spkt = np.array(timestamps)[spike_train_frames]
                else:
                    spkt = sorting.get_unit_spike_train(unit_id=unit_id) / fs
                for pr in all_properties:
                    if pr not in skip_properties:
                        if pr in sorting.get_unit_property_names(unit_id):
                            prop_value = sorting.get_unit_property(unit_id, pr)
                            unit_kwargs.update({pr: prop_value})
                        else:  # Case of missing data for this unit and this property
                            unit_kwargs.update({pr: np.nan})
                nwbfile.add_unit(id=unit_id, spike_times=spkt, **unit_kwargs)

            # TODO
            # # Stores average and std of spike traces
            # This will soon be updated to the current NWB standard
            # if 'waveforms' in sorting.get_unit_spike_feature_names(unit_id=id):
            #     wf = sorting.get_unit_spike_features(unit_id=id,
            #                                          feature_name='waveforms')
            #     relevant_ch = most_relevant_ch(wf)
            #     # Spike traces on the most relevant channel
            #     traces = wf[:, relevant_ch, :]
            #     traces_avg = np.mean(traces, axis=0)
            #     traces_std = np.std(traces, axis=0)
            #     nwbfile.add_unit(
            #         id=id,
            #         spike_times=spkt,
            #         waveform_mean=traces_avg,
            #         waveform_sd=traces_std
            #     )

            # check that multidimensional features have the same shape across units
            feature_shapes = {}
            skip_features = []
            for ft in all_features:
                shapes = []
                for unit_id in unit_ids:
                    if ft in sorting.get_unit_spike_feature_names(unit_id):
                        feat_value = sorting.get_unit_spike_features(unit_id, ft)
                        if isinstance(feat_value[0], (int, np.integer, float, np.float, str, bool)):
                            break
                        elif isinstance(feat_value[0], (list, np.ndarray)):  # multidimensional features
                            if np.array(feat_value).ndim > 1:
                                shapes.append(np.array(feat_value).shape)
                                feature_shapes[ft] = shapes
                        elif isinstance(feat_value[0], dict):
                            print(f"Skipping feature '{ft}' because dictionaries are not supported.")
                            skip_features.append(ft)
                            break
                    else:
                        print(f"Skipping feature '{ft}' because not share across all units.")
                        skip_features.append(ft)
                        break

            nspikes = {k: get_nspikes(nwbfile.units, int(k)) for k in unit_ids}

            for ft in feature_shapes.keys():
                # skip first dimension (num_spikes) when comparing feature shape
                if not np.all([elem[1:] == feature_shapes[ft][0][1:] for elem in feature_shapes[ft]]):
                    print(f"Skipping feature '{ft}' because it has variable size across units.")
                    skip_features.append(ft)

            for ft in all_features:
                values = []
                if not ft.endswith('_idxs'):
                    for unit_id in sorting.get_unit_ids():
                        feat_vals = sorting.get_unit_spike_features(unit_id, ft)

                        if len(feat_vals) < nspikes[unit_id]:
                            skip_features.append(ft)
                            print(f"Skipping feature '{ft}' because it is not defined for all spikes.")
                            break
                            # this means features are available for a subset of spikes
                            # all_feat_vals = np.array([np.nan] * nspikes[unit_id])
                            # feature_idxs = sorting.get_unit_spike_features(unit_id, feat_name + '_idxs')
                            # all_feat_vals[feature_idxs] = feat_vals
                        else:
                            all_feat_vals = feat_vals
                        values.append(all_feat_vals)

                    if ft not in skip_features:
                        flatten_vals = [item for sublist in values for item in sublist]
                        nspks_list = [sp for sp in nspikes.values()]
                        spikes_index = np.cumsum(nspks_list).astype('int64')

                        set_dynamic_table_property(
                            dynamic_table=nwbfile.units,
                            row_ids=unit_ids,
                            property_name=ft,
                            values=flatten_vals,
                            index=spikes_index,
                        )

        else:  # there are already units in the nwbfile
            print("Warning: The nwbfile already contains units. "
                  "These units will not be over-written.")

    @staticmethod
    def write_sorting(sorting: se.SortingExtractor, save_path: PathType = None, nwbfile=None,
                      property_descriptions: dict = None, timestamps: ArrayType = None, **nwbfile_kwargs):
        '''
        Parameters
        ----------
        sorting: SortingExtractor
        save_path: PathType
            Required if an nwbfile is not passed. Must be the path to the nwbfile
            being appended, otherwise one is created and written.
        nwbfile: NWBFile
            Required if a save_path is not specified. If passed, this function
            will fill the relevant fields within the nwbfile. E.g., calling
            spikeextractors.NwbRecordingExtractor.write_recording(
                my_recording_extractor, my_nwbfile
            )
            will result in the appropriate changes to the my_nwbfile object.
        property_descriptions: dict
            For each key in this dictionary which matches the name of a unit
            property in sorting, adds the value as a description to that
            custom unit column.
        timestamps: array-like
            If provided, the timestamps in seconds or the assiciated RecordingExtractor to be saved as the unit
            timestamps. (default=None)
        nwbfile_kwargs: dict
            Information for constructing the nwb file (optional).
            Only used if no nwbfile exists at the save_path, and no nwbfile
            was directly passed.
        '''
        assert HAVE_NWB, NwbSortingExtractor.installation_mesg
        assert save_path is None or nwbfile is None, \
            "Either pass a save_path location, or nwbfile object, but not both!"

        if nwbfile is None:
            if Path(save_path).is_file():
                read_mode = 'r+'
            else:
                read_mode = 'w'

            with NWBHDF5IO(save_path, mode=read_mode) as io:
                if read_mode == 'r+':
                    nwbfile = io.read()
                else:
                    kwargs = {'session_description': 'No description',
                              'identifier': str(uuid.uuid4()),
                              'session_start_time': datetime.now()}
                    kwargs.update(**nwbfile_kwargs)
                    nwbfile = NWBFile(**kwargs)
                se.NwbSortingExtractor.write_units(
                    sorting=sorting,
                    nwbfile=nwbfile,
                    property_descriptions=property_descriptions,
                    timestamps=timestamps
                )

                io.write(nwbfile)
        else:
            assert isinstance(nwbfile, NWBFile), "'nwbfile' should be of type pynwb.NWBFile"
            se.NwbSortingExtractor.write_units(
                    sorting=sorting,
                    nwbfile=nwbfile,
                    property_descriptions=property_descriptions,
                    timestamps=timestamps
            )



import kachery_client as kc

# @kc.taskfunction('sortingview.get_figure_object.1', type='pure-calculation')
def task_get_figure_object(figure_object_hash: str):
    # TODO: only load from local kachery storage and validate data prior to returning
    obj = kc.load_json('sha1://' + figure_object_hash)
    return obj


from typing import Union
import kachery_client as kc

key = '_sortingview_user_permissions'

def set_user_permissions(user_id: str, *, append_to_all_feeds: Union[None, bool]=None):
    p = kc.get(key, {})
    p_user = p.get(user_id, {})
    if append_to_all_feeds is not None:
        p_user['appendToAllFeeds'] = append_to_all_feeds
    p[user_id] = p_user
    kc.set(key, p)

def set_user_feed_permissions(user_id: str, *, feed_id: str, append: Union[None, bool]=None):
    p = kc.get(key)
    if p is None: p = {}
    p_user = p.get(user_id, {})
    feeds = p_user.get('feeds', {})
    feed = feeds.get(feed_id, {})
    if append is not None:
        feed['append'] = append
    feeds[feed_id] = feed
    p_user['feeds'] = feeds
    p[user_id] = p_user
    kc.set(key, p)

def get_user_permissions_dict(user_id: str):
    p = kc.get(key)
    if p is None: p = {}
    p_user = p.get(user_id, {})
    return p_user


import numpy as np

def _estimate_sampling_frequency(timestamps: np.ndarray):
    if len(timestamps) <= 1:
        return 0
    deltas = np.diff(timestamps)
    median_delta = np.median(deltas)
    deltas_excluding_outliers = deltas[(median_delta * 0.9 < deltas) & (deltas < median_delta * 1.1)]
    return 1 / float(np.mean(deltas_excluding_outliers))



"""
Code ported from cortex-lab/phylib module
"""

import numpy as np


def _index_of(arr, lookup):
    """Replace scalars in an array by their indices in a lookup table.
    Implicitely assume that:
    * All elements of arr and lookup are non-negative integers.
    * All elements or arr belong to lookup.
    This is not checked for performance reasons.
    """
    # Equivalent of np.digitize(arr, lookup) - 1, but much faster.
    # TODO: assertions to disable in production for performance reasons.
    # TODO: np.searchsorted(lookup, arr) is faster on small arrays with large
    # values
    lookup = np.asarray(lookup, dtype=np.int32)
    m = (lookup.max() if len(lookup) else 0) + 1
    tmp = np.zeros(m + 1, dtype=np.int)
    # Ensure that -1 values are kept.
    tmp[-1] = -1
    if len(lookup):
        tmp[lookup] = np.arange(len(lookup))
    return tmp[arr]


def _increment(arr, indices):
    """Increment some indices in a 1D vector of non-negative integers.
    Repeated indices are taken into account."""
    bbins = np.bincount(indices)
    arr[:len(bbins)] += bbins
    return arr


def _diff_shifted(arr, steps=1):
    return arr[steps:] - arr[:len(arr) - steps]


def _create_correlograms_array(n_clusters, winsize_bins):
    return np.zeros((n_clusters, n_clusters, winsize_bins // 2 + 1),
                    dtype=np.int32)


def _symmetrize_correlograms(correlograms):
    """Return the symmetrized version of the CCG arrays."""

    n_clusters, _, n_bins = correlograms.shape
    assert n_clusters == _

    # We symmetrize c[i, j, 0].
    # This is necessary because the algorithm in correlograms()
    # is sensitive to the order of identical spikes.
    correlograms[..., 0] = np.maximum(correlograms[..., 0],
                                      correlograms[..., 0].T)

    sym = correlograms[..., 1:][..., ::-1]
    sym = np.transpose(sym, (1, 0, 2))

    return np.dstack((sym, correlograms))


def compute_correlograms(spike_times, spike_clusters, cluster_ids=None, sample_rate=1.,
                         bin_size=None, window_size=None, symmetrize=True):
    """Compute all pairwise cross-correlograms among the clusters appearing
    in `spike_clusters`.
    Parameters
    ----------
    spike_times : array-like
        Spike times in seconds.
    spike_clusters : array-like
        Spike-cluster mapping.
    cluster_ids : array-like
        The list of *all* unique clusters, in any order. That order will be used
        in the output array.
    bin_size : float
        Size of the bin, in seconds.
    window_size : float
        Size of the window, in seconds.
    sample_rate : float
        Sampling rate.
    symmetrize : boolean (True)
        Whether the output matrix should be symmetrized or not.
    Returns
    -------
    correlograms : array
        A `(n_clusters, n_clusters, winsize_samples)` array with all pairwise CCGs.
    """
    assert sample_rate > 0.
    assert np.all(np.diff(spike_times) >= 0), ("The spike times must be "
                                               "increasing.")

    # Get the spike samples.
    spike_times = np.asarray(spike_times, dtype=np.float64)
    spike_samples = (spike_times * sample_rate).astype(np.int64)

    assert spike_samples.ndim == 1
    assert spike_samples.shape == spike_clusters.shape

    # Find `binsize`.
    bin_size = np.clip(bin_size, 1e-5, 1e5)  # in seconds
    binsize = int(sample_rate * bin_size)  # in samples
    assert binsize >= 1

    # Find `winsize_bins`.
    window_size = np.clip(window_size, 1e-5, 1e5)  # in seconds
    winsize_bins = 2 * int(.5 * window_size / bin_size) + 1

    assert winsize_bins >= 1
    assert winsize_bins % 2 == 1

    # Take the cluster order into account.
    if cluster_ids is None:
        clusters = np.unique(spike_clusters)
    else:
        clusters = np.array(cluster_ids)
    n_clusters = len(clusters)

    # Like spike_clusters, but with 0..n_clusters-1 indices.
    # spike_clusters_i = _index_of(spike_clusters, clusters)
    spike_clusters_i = _index_of(spike_clusters, clusters)

    # Shift between the two copies of the spike trains.
    shift = 1

    # At a given shift, the mask precises which spikes have matching spikes
    # within the correlogram time window.
    mask = np.ones_like(spike_samples, dtype=np.bool)

    correlograms = _create_correlograms_array(n_clusters, winsize_bins)

    # The loop continues as long as there is at least one spike with
    # a matching spike.
    while mask[:-shift].any():
        # Number of time samples between spike i and spike i+shift.
        spike_diff = _diff_shifted(spike_samples, shift)

        # Binarize the delays between spike i and spike i+shift.
        spike_diff_b = spike_diff // binsize

        # Spikes with no matching spikes are masked.
        mask[:-shift][spike_diff_b > (winsize_bins // 2)] = False

        # Cache the masked spike delays.
        m = mask[:-shift].copy()
        d = spike_diff_b[m]

        # # Update the masks given the clusters to update.
        # m0 = np.in1d(spike_clusters[:-shift], clusters)
        # m = m & m0
        # d = spike_diff_b[m]
        d = spike_diff_b[m]

        # Find the indices in the raveled correlograms array that need
        # to be incremented, taking into account the spike clusters.
        indices = np.ravel_multi_index(
            (spike_clusters_i[:-shift][m], spike_clusters_i[+shift:][m], d), correlograms.shape)

        # Increment the matching spikes in the correlograms array.
        _increment(correlograms.ravel(), indices)

        shift += 1

    # Remove ACG peaks. -- exclude because this is wrong (jfm)
    # correlograms[np.arange(n_clusters), np.arange(n_clusters), 0] = 0

    if symmetrize:
        return _symmetrize_correlograms(correlograms)
    else:
        return correlograms



import numpy as np
import spikeextractors as se
from .get_unit_waveforms import get_unit_waveforms
from .SubsampledSortingExtractor import SubsampledSortingExtractor

def find_unit_peak_channels(recording, sorting, unit_ids):
    # Use the first part of the recording to estimate the peak channels
    sorting_shortened = SubsampledSortingExtractor(parent_sorting=sorting, max_events_per_unit=20, method='truncate')
    max_time = 0
    for unit_id in sorting_shortened.get_unit_ids():
        st = sorting_shortened.get_unit_spike_train(unit_id=unit_id)
        if len(st) > 0:
            max_time = max(max_time, np.max(st))
    recording_shortened = se.SubRecordingExtractor(parent_recording=recording, start_frame=0, end_frame=max_time + 1)
    unit_waveforms = get_unit_waveforms(
        recording=recording_shortened,
        sorting=sorting_shortened,
        unit_ids=unit_ids,
        channel_ids_by_unit=None,
        snippet_len=(10, 10)
    )
    channel_ids = recording.get_channel_ids()
    peak_channels = {}
    for ii, unit_id in enumerate(unit_ids):
        average_waveform = np.median(unit_waveforms[ii], axis=0)
        peak_channel_index = int(np.argmax(np.max(average_waveform, axis=1) - np.min(average_waveform, axis=1)))
        peak_channels[unit_id] = int(channel_ids[peak_channel_index])
    return peak_channels



from .labboxephysrecordingextractor import LabboxEphysRecordingExtractor
from .labboxephyssortingextractor import LabboxEphysSortingExtractor
from .mdaextractors import writemda32
from .mdaextractors import MdaRecordingExtractor, MdaSortingExtractor
from .h5extractors.h5sortingextractorv1 import H5SortingExtractorV1


from typing import Union
import numpy as np
import spikeextractors as se

class SubsampledSortingExtractor(se.SortingExtractor):
    # method: 'random', 'truncate'
    def __init__(self, parent_sorting, *, max_events_per_unit: Union[int, None]=None, method='random'):
        se.SortingExtractor.__init__(self)
        self._parent_sorting = parent_sorting
        self._subsampled_spike_trains_by_unit = {}
        unit_ids = parent_sorting.get_unit_ids()
        for unit_id in unit_ids:
            if max_events_per_unit is not None:
                st = parent_sorting.get_unit_spike_train(unit_id=unit_id)
                if len(st) > max_events_per_unit:
                    if method == 'random':
                        indices = np.sort(np.random.RandomState(seed=0).choice(np.arange(len(st)), size=max_events_per_unit, replace=False))
                    elif method == 'truncate':
                        indices = np.arange(max_events_per_unit)
                    else:
                        raise Exception('Unexpected method in SubsampledSortingExtractor')
                    self._subsampled_spike_trains_by_unit[int(unit_id)] = st[indices]
                else:
                    self._subsampled_spike_trains_by_unit[int(unit_id)] = None
            else:
                self._subsampled_spike_trains_by_unit[int(unit_id)] = None
        
    def make_serialized_dict(self):
        # this is necessary because the base class implementation
        # seems to have a bug where it asks for imported_module.__version__, which may not be defined
        return None

    def get_unit_ids(self):
        return self._parent_sorting.get_unit_ids()

    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        st2 = self._subsampled_spike_trains_by_unit[int(unit_id)]
        if st2 is None:
            return self._parent_sorting.get_unit_spike_train(unit_id=unit_id, start_frame=start_frame, end_frame=end_frame)
        if start_frame is None and end_frame is None:
            return st2
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = np.Inf
        return st2[(start_frame <= st2) & (st2 < end_frame)]

    def get_sampling_frequency(self):
        return self._parent_sorting.get_sampling_frequency()


# import h5py
from typing import List
import warnings

import numpy as np
from spikeextractors import SortingExtractor

warnings.simplefilter(action='ignore', category=FutureWarning)
import h5py

warnings.resetwarnings()

class H5SortingExtractorV1(SortingExtractor):
    def __init__(self, h5_path):
        SortingExtractor.__init__(self)
        self._h5_path = h5_path
        self._loaded_spike_trains = {}

        with h5py.File(self._h5_path, 'r') as f:
            self._unit_ids = np.array(f.get('unit_ids'))
            self._sampling_frequency = np.array(f.get('sampling_frequency'))[0].item()
            if np.isnan(self._sampling_frequency):
                print('WARNING: sampling frequency is nan. Using 30000 for now. Please correct the snippets file.')
                self._sampling_frequency = 30000

    def get_unit_ids(self):
        return self._unit_ids.tolist()

    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        if unit_id not in self._loaded_spike_trains:
            with h5py.File(self._h5_path, 'r') as f:
                self._loaded_spike_trains[unit_id] = np.array(f.get(f'unit_spike_trains/{unit_id}'))

        x = self._loaded_spike_trains[unit_id]

        if start_frame is None:
            start_frame = 0
        
        if (start_frame == 0) and (end_frame is None):
            return x
        
        assert end_frame is not None

        return np.array([t for t in x if ((start_frame <= t) and (t < end_frame))])

    def get_sampling_frequency(self):
        return self._sampling_frequency

    @staticmethod
    def write_sorting(sorting, save_path):
        unit_ids = sorting.get_unit_ids()
        samplerate = sorting.get_sampling_frequency()
        W = H5SortingExtractorV1Writer(save_path=save_path, samplerate=samplerate)
        for unit_id in unit_ids:
            W.add_unit(unit_id=unit_id, times=sorting.get_unit_spike_train(unit_id=unit_id))
        W.finalize()

class H5SortingExtractorV1Writer:
    def __init__(self, *, save_path: str, samplerate: float):
        self._save_path = save_path
        self._samplerate = samplerate
        self._unit_ids: List[int] = []
        with h5py.File(save_path, 'w') as f:
            f.create_dataset('sampling_frequency', data=np.array([samplerate]).astype(np.float64))
    def add_unit(self, *, unit_id: int, times: np.array):
        with h5py.File(self._save_path, 'a') as f:
            f.create_dataset(f'unit_spike_trains/{unit_id}', data=np.array(times).astype(np.int64))
        self._unit_ids.append(unit_id)
    def finalize(self):
        with h5py.File(self._save_path, 'a') as f:
            f.create_dataset('unit_ids', data=np.array(self._unit_ids).astype(np.int32))


from typing import List
import numpy as np
import kachery_client as kc
from figurl import Figure

class SpikeExplorer:
    def __init__(
        self,
        snippets: np.ndarray, # L x T x M
        timestamps: np.array, #L
        features: np.ndarray, # L x K
        feature_names: List[str]
    ) -> None:
        self._snippets = snippets
        self._timestamps = timestamps
        self._features = features
        self._feature_names = feature_names
    def figurl(self):
        data = {
            'numSpikes': len(self._timestamps),
            'numChannels': int(self._snippets.shape[2]),
            'featureNames': self._feature_names,
            'snippetsUri': kc.store_npy(self._snippets.astype(np.float32)),
            'timestampsUri': kc.store_npy(self._timestamps.astype(np.float32)),
            'featuresUri': kc.store_npy(self._features.astype(np.float32))
        }
        return Figure(type='sortingview.spikeexplorer.1', data=data)


from hdmf.utils import docval, call_docval_func, get_docval
from pynwb import register_class, NWBContainer
from pynwb.core import NWBDataInterface, MultiContainerInterface
from pynwb.device import Device


@register_class('ShanksElectrode', 'ndx-franklab-novela')
class ShanksElectrode(NWBContainer):
    ''' Representation of ShanksElectrode object in NWB '''

    @docval(
        {'name': 'name', 'type': str, 'doc': 'name of the shank'},
        {'name': 'rel_x', 'type': float, 'doc': 'the rel_x value of this electrode'},
        {'name': 'rel_y', 'type': float, 'doc': 'the rel_y value of this electrode'},
        {'name': 'rel_z', 'type': float, 'doc': 'the rel_z value of this electrode'},
    )
    def __init__(self, **kwargs):
        super(ShanksElectrode, self).__init__(name=kwargs['name'])
        self.rel_x = kwargs['rel_x']
        self.rel_y = kwargs['rel_y']
        self.rel_z = kwargs['rel_z']


@register_class('Shank', 'ndx-franklab-novela')
class Shank(MultiContainerInterface):
    ''' Representation of Shank object in NWB '''

    @docval(
        {'name': 'name', 'type': str, 'doc': 'name of the shank'},
        {'name': 'shanks_electrodes', 'type': (list, tuple), 'doc': 'electrodes in shank', 'default': list()}
    )
    def __init__(self, **kwargs):
        super(Shank, self).__init__(name=kwargs['name'])
        self.shanks_electrodes = kwargs['shanks_electrodes']

    __clsconf__ = [
        {
            'attr': 'shanks_electrodes',
            'type': ShanksElectrode,
            'add': 'add_shanks_electrode',
            'get': 'get_shanks_electrode'
        }
    ]


@register_class('Probe', 'ndx-franklab-novela')
class Probe(Device, MultiContainerInterface):
    ''' Representation of Probe object in NWB '''

    @docval(*get_docval(Device.__init__) + (
            {'name': 'id', 'type': 'int', 'doc': 'unique id of the probe'},
            {'name': 'probe_type', 'type': 'str', 'doc': 'type of probe'},
            {'name': 'units', 'type': 'str', 'doc': 'units in device'},
            {'name': 'probe_description', 'type': 'str', 'doc': 'description of the probe'},
            {'name': 'contact_side_numbering', 'type': 'bool', 'doc': 'is contact_side_numbering enabled'},
            {'name': 'contact_size', 'type': 'float', 'doc': 'value of contact size as float'},
            {'name': 'shanks', 'type': (list, tuple), 'doc': 'shanks in probe', 'default': list()}
    ))
    def __init__(self, **kwargs):
        super().__init__(**{kwargs_item: kwargs[kwargs_item]
                            for kwargs_item in kwargs.copy()
                            if kwargs_item != 'id'
                            if kwargs_item != 'probe_type'
                            if kwargs_item != 'units'
                            if kwargs_item != 'probe_description'
                            if kwargs_item != 'contact_side_numbering'
                            if kwargs_item != 'contact_size'
                            if kwargs_item != 'shanks'
                            })
        call_docval_func(super(Probe, self).__init__, kwargs)
        self.id = kwargs['id']
        self.probe_type = kwargs['probe_type']
        self.units = kwargs['units']
        self.probe_description = kwargs['probe_description']
        self.contact_side_numbering = kwargs['contact_side_numbering']
        self.contact_size = kwargs['contact_size']
        self.shanks = kwargs['shanks']

    __nwbfields__ = ('id', 'probe_type', 'units', 'probe_description', 'contact_side_numbering',
                     'contact_size')

    __clsconf__ = [
        {
            'attr': 'shanks',
            'type': Shank,
            'add': 'add_shank',
            'get': 'get_shank'
        }
    ]



from typing import Any
import kachery_client as kc

# @kc.taskfunction('latency_test_query.1', type='query')
def task_latency_test_query(x: str):
    return x

# @kc.taskfunction('get_action_latency_test_subfeed.1', type='query')
def get_action_latency_test_subfeed():
    feed = kc.load_feed('_latency_test', create=True)
    subfeed = feed.load_subfeed('main')
    return {
        'feedId': feed.feed_id,
        'subfeedHash': subfeed.subfeed_hash
    }

# @kc.taskfunction('subfeed_latency_test_append.1', type='query')
def subfeed_latency_test_append(message: Any):
    feed = kc.load_feed('_latency_test')
    subfeed = feed.load_subfeed('main')
    subfeed.append_message(message)
    return {'numMessages': subfeed.get_num_local_messages()}


import kachery_client as kc
from sortingview.workspace import Workspace
from ..workspace import Workspace

# Obsolete, but needed for now for the migration script to work
class WorkspaceList:
    def __init__(self, *, list_name: str):
        print(f'WARNING: WorkspaceList is obsolete (except when migrating to new system). Instead, use get_workspace_list(), set_workspace_list(), and add_workspace_to_list().')
        self._list_name = list_name
        self._workspaces = {}
        sf = kc.load_subfeed(self.get_subfeed_uri())
        messages = sf.get_next_messages()
        for m in messages:
            a = m.get('action', {})
            type0 = a.get('type', None)
            if type0 == 'add':
                w = a.get('workspace')
                self._workspaces[w['name']] = w
            elif type0 == 'remove':
                name = a.get('workspaceName')
                if name in self._workspaces:
                    del self._workspaces[name]
    @property
    def workspace_names(self):
        return [name for name in list(self._workspaces.keys())]
    def get_workspace(self, name):
        w = self._workspaces[name]
        return Workspace(workspace_uri=w['uri'], label=w['name'])
    def add_workspace(self, *, name: str, workspace: Workspace):
        if name in self.workspace_names:
            raise Exception(f'Workspace with name already exists: {name}')
        sf = kc.load_subfeed(self.get_subfeed_uri())
        w = {
            'name': name,
            'uri': workspace.uri
        }
        sf.append_message({
            'action': {
                'type': 'add',
                'workspace': w
            }
        })
        self._workspaces[name] = w
    def remove_workspace(self, name: str):
        if name not in self.workspace_names:
            raise Exception(f'Workspace not found: {name}')
        sf = kc.load_subfeed(self.get_subfeed_uri())
        sf.append_message({
            'action': {
                'type': 'remove',
                'workspaceName': name
            }
        })
        del self._workspaces[name]
    def get_subfeed_uri(self):
        k = {'name': 'sortingview-workspace-list'}
        feed_uri = kc.get(k)
        if feed_uri is None:
            feed_uri = kc.create_feed().uri
            kc.set(k, feed_uri)
        f = kc.load_feed(feed_uri)
        return f.load_subfeed(self._list_name).uri


import os
import hither2 as hi
import kachery_client as kc
from sortingview.config import job_cache, job_handler

@hi.function(
    'get_firing_data', '0.1.2',
    image=hi.RemoteDockerImage('docker://magland/labbox-ephys-processing:0.3.19'),
    modules=['sortingview']
)
def get_firing_data(sorting_object, recording_object, configuration={}, snippet_len=(50, 80)):
    from decimal import Decimal
    S, R = get_structure(sorting_object, recording_object)
    elapsed = R.get_num_frames()/R.get_sampling_frequency()
    ids = S.get_unit_ids()
    train = [S.get_unit_spike_train(id).size for id in ids]
    keyedCount = dict(zip(
        [str(id) for id in ids],
        [{'count': t,
          'rate': f"{Decimal(t / elapsed).quantize(Decimal('.01'))}"} for t in train]))
    return keyedCount

@kc.taskfunction('get_firing_data.1', type='pure-calculation')
def task_get_firing_data(sorting_object, recording_object, configuration={}, snippet_len=(50, 80)):
    with hi.Config(
        job_cache=job_cache,
        job_handler=job_handler.metrics
    ):
        return get_firing_data.run(
            sorting_object=sorting_object,
            recording_object=recording_object,
            configuration=configuration
        )


def get_structure(sorting_object, recording_object):
    from sortingview.extractors import LabboxEphysRecordingExtractor, LabboxEphysSortingExtractor
    S = LabboxEphysSortingExtractor(sorting_object)
    R = LabboxEphysRecordingExtractor(recording_object)
    return S, R




from typing import Dict, List, cast
import numpy as np
import spikeextractors as se

class CuratedSortingExtractor(se.SortingExtractor):
    def __init__(self, *, parent_sorting: se.SortingExtractor, merge_groups: List[List[int]]):
        se.SortingExtractor.__init__(self)
        self._parent_sorting = parent_sorting
        self._merge_groups = merge_groups
        uids = cast(List[int], parent_sorting.get_unit_ids())
        self._unit_mapping: Dict[int, int] = {}
        self._merge_groups_by_unit_id: Dict[int, List[int]] = {}
        for u in uids:
            self._unit_mapping[u] = u
            self._merge_groups_by_unit_id[u]= [u]
        for mg in merge_groups:
            minval = min(mg)
            for x in mg:
                self._unit_mapping[x] = minval
                self._merge_groups_by_unit_id[x] = mg
        self._unit_ids = [u for u in uids if (self._unit_mapping[u] == u)]
        self._sampling_frequency = parent_sorting.get_sampling_frequency()

    def get_unit_ids(self):
        return self._unit_ids

    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        mg = self._merge_groups_by_unit_id[unit_id]
        trains = [self._parent_sorting.get_unit_spike_train(unit_id=x) for x in mg]
        if len(trains) == 1:
            return trains[0]
        else:
            return np.sort(np.concatenate(trains))

    def get_sampling_frequency(self):
        return self._sampling_frequency



import os
import hither2 as hi
import kachery_client as kc
import numpy as np
from sortingview.config import job_cache, job_handler

from sortingview.helpers import prepare_snippets_h5
from sortingview.helpers.get_unit_waveforms import get_unit_waveforms

@hi.function(
    'get_unit_snrs', '0.1.0',
    image=hi.RemoteDockerImage('docker://magland/labbox-ephys-processing:0.3.19'),
    modules=['sortingview']
)
def get_unit_snrs(snippets_h5):
    import h5py
    h5_path = kc.load_file(snippets_h5)
    assert h5_path is not None
    ret = {}
    with h5py.File(h5_path, 'r') as f:
        unit_ids = np.array(f.get('unit_ids'))
        for unit_id in unit_ids:
            unit_waveforms = np.array(f.get(f'unit_waveforms/{unit_id}/waveforms')) # n x M x T
            ret[str(unit_id)] = _compute_unit_snr_from_waveforms(unit_waveforms)
    return ret

def _compute_unit_snr_from_waveforms(waveforms):
    average_waveform = np.mean(waveforms, axis=0)
    channel_amplitudes = (np.max(average_waveform, axis=1) - np.min(average_waveform, axis=1)).squeeze() # M
    peak_channel_index = np.argmax(channel_amplitudes)
    mean_subtracted_waveforms_on_peak_channel = waveforms[:, peak_channel_index, :] - average_waveform[peak_channel_index]
    est_noise_level = np.median(np.abs(mean_subtracted_waveforms_on_peak_channel.squeeze())) / 0.6745  # median absolute deviation (MAD) estimate of stdev
    peak_channel_amplitude = channel_amplitudes[peak_channel_index]
    snr = peak_channel_amplitude / est_noise_level
    return snr

@kc.taskfunction('get_unit_snrs.1', type='pure-calculation')
def task_get_unit_snrs(sorting_object, recording_object, configuration={}, snippet_len=(50, 80)):
    with hi.Config(
        job_cache=job_cache,
        job_handler=job_handler.metrics
    ):
        with hi.Config(job_handler=job_handler.extract_snippets):
            snippets_h5 = prepare_snippets_h5.run(recording_object=recording_object, sorting_object=sorting_object, snippet_len=snippet_len)
        return get_unit_snrs.run(
            snippets_h5=snippets_h5
        )


from typing import List, cast
import hither2 as hi
import numpy as np
import spikeextractors as se
import kachery_client as kc
from sortingview.config import job_cache, job_handler
from sortingview.extractors import LabboxEphysSortingExtractor, LabboxEphysRecordingExtractor

@hi.function(
    'recording_info', '0.1.3'
)
def recording_info(recording_uri):
    recording = LabboxEphysRecordingExtractor(recording_uri, download=False)
    return dict(
        sampling_frequency=recording.get_sampling_frequency(),
        channel_ids=recording.get_channel_ids(),
        channel_groups=recording.get_channel_groups().tolist(),
        geom=geom_from_recording(recording).tolist(),
        num_frames=recording.get_num_frames(),
        noise_level=estimate_noise_level(recording),
        recording_object=recording.object()
    )

@kc.taskfunction('recording_info.3', type='pure-calculation')
def task_recording_info(recording_uri: str):
    with hi.Config(job_handler=job_handler.misc, job_cache=job_cache):
        return hi.Job(recording_info, {'recording_uri': recording_uri})

def geom_from_recording(recording):
    channel_ids = recording.get_channel_ids()
    location0 = recording.get_channel_property(channel_ids[0], 'location')
    nd = len(location0)
    M = len(channel_ids)
    geom = np.zeros((M, nd))
    for ii in range(len(channel_ids)):
        location_ii = recording.get_channel_property(channel_ids[ii], 'location')
        geom[ii, :] = list(location_ii)
    return geom

def estimate_noise_level(recording: se.RecordingExtractor):
    N = cast(int, recording.get_num_frames())
    samplerate = cast(float, recording.get_sampling_frequency())
    start_frame = 0
    end_frame = np.minimum(int(samplerate), N)
    X = cast(np.ndarray, recording.get_traces(channel_ids=[int(id) for id in cast(List[int], recording.get_channel_ids())], start_frame=start_frame, end_frame=end_frame))
    X_mean_subtracted = _mean_subtract_on_channels(X)
    est_noise_level = np.median(np.abs(X_mean_subtracted.squeeze())) / 0.6745  # median absolute deviation (MAD) estimate of stdev
    if (est_noise_level == 0): est_noise_level = 1
    return est_noise_level

def _mean_subtract_on_channels(X: np.ndarray):
    return X - np.broadcast_to(np.mean(X, axis=1), (X.shape[1], X.shape[0])).T


from typing import Any
import kachery_client as kc
from ..workspace import Workspace

def assert_item_is_valid_workspace_entry(w: Any) -> None:
    if not isinstance(w, dict): raise Exception('Invalid workspace: not a dictionary')
    if not isinstance(w.get('workspaceUri', None), str): raise Exception('Missing or invalid field in workspace: workspaceUri')
    if not isinstance(w.get('label', None), str): raise Exception('Missing or invalid field in workspace: label')
    if not isinstance(w.get('metaData', {}), dict): raise Exception('Invalid field in workspace: metaData')
    for k in w.keys():
        if k not in ['workspaceUri', 'label', 'metaData']:
            raise Exception(f'Invalid field in workspace: {k}')

def get_workspace_list(name: str='default'):
    x = kc.get({'type': 'sortingview-workspace-list', 'name': name})
    if x is None: x = []
    return x

def set_workspace_list(workspace_list, *, name: str='default'):
    if not isinstance(workspace_list, list): raise Exception('Not a list')
    map(assert_item_is_valid_workspace_entry, workspace_list)
    kc.set({'type': 'sortingview-workspace-list', 'name': name}, workspace_list)

def add_workspace_to_list(*, list_name: str='default', workspace: Workspace, meta_data: Any={}):
    x = get_workspace_list(name=list_name)
    x.append({
        'workspaceUri': workspace.uri,
        'label': workspace.label,
        'metaData': meta_data
    })
    set_workspace_list(x)

# TODO: remove workspace from list


from os import stat
from typing import List, Set

import h5py
import kachery_client as kc
import numpy as np
import spikeextractors as se

from .TimeseriesModel_Hdf5.TimeseriesModel_Hdf5 import TimeseriesModel_Hdf5, prepare_timeseries_hdf5_from_recording, set_geom_on_recording


class H5RecordingExtractorV1(se.RecordingExtractor):
    extractor_name = 'H5RecordingExtractorV1'
    is_writable = False
    def __init__(self, *, h5_path: str):
        se.RecordingExtractor.__init__(self)
        
        self._h5_path: str = h5_path
        self._timeseries_model = TimeseriesModel_Hdf5(self._h5_path)
        geom = self._timeseries_model.geom()
        if geom is not None:
            set_geom_on_recording(self, geom)

    def get_channel_ids(self) -> List[int]:
        return self._timeseries_model.channelIds()

    def get_num_frames(self) -> int:
        return self._timeseries_model.numTimepoints()

    def get_sampling_frequency(self) -> float:
        return self._timeseries_model.samplingFrequency()

    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None, return_scaled=True):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = self.get_num_frames()
        if channel_ids is None:
            channel_ids = self.get_channel_ids()
        M = len(channel_ids)
        N = end_frame - start_frame

        all_channel_ids = self.get_channel_ids()
        channel_inds = [all_channel_ids.index(id) for id in channel_ids]

        return self._timeseries_model.getChunk(t1=start_frame, t2=end_frame, channel_inds=channel_inds)
    @staticmethod
    def write_recording(recording: se.RecordingExtractor, h5_path: str, dtype=float):
        hdf5_chunk_size = 10 * 1000 * 1000
        hdf5_padding = 10000
        prepare_timeseries_hdf5_from_recording(recording, h5_path, chunk_size=hdf5_chunk_size, padding=hdf5_padding, dtype=dtype)




from hdmf.utils import docval, call_docval_func, get_docval
from pynwb import register_class
from pynwb.device import Device


@register_class('HeaderDevice', 'ndx-franklab-novela')
class HeaderDevice(Device):
    """Configuration comes from SpikeGadgets recording system. Represented as HeaderDevice in NWB"""

    __nwbfields__ = ('headstage_serial', 'headstage_smart_ref_on', 'realtime_mode', 'headstage_auto_settle_on',
                     'timestamp_at_creation', 'conntroller_firmware_version', 'controller_serial',
                     'save_displayed_chan_only', 'headstage_firmwavare_version', 'qt_version', 'compile_date',
                     'compile_time', 'file_prefix', 'headstage_gyro_sensor_on', 'headstage_mag_sensor_on',
                     'trodes_version', 'headstage_accel_sensor_on', 'commit_head', 'system_time_at_creation',
                     'file_path')

    @docval(*get_docval(Device.__init__) + (
            {'name': 'headstage_serial', 'type': 'str', 'doc': 'headstage_serial from header global configuration'},
            {'name': 'headstage_smart_ref_on', 'type': 'str', 'doc': 'headstage_smart_ref_on from header global configuration'},
            {'name': 'realtime_mode', 'type': 'str', 'doc': 'realtime_mode from header global configuration'},
            {'name': 'headstage_auto_settle_on', 'type': 'str', 'doc': 'headstage_auto_settle_on from header global configuration'},
            {'name': 'timestamp_at_creation', 'type': 'str', 'doc': 'timestamp_at_creation from header global configuration'},
            {'name': 'controller_firmware_version', 'type': 'str', 'doc': 'conntroller_firmware_version from header global configuration'},
            {'name': 'controller_serial', 'type': 'str', 'doc': 'conntroller_serial from header global configuration'},
            {'name': 'save_displayed_chan_only', 'type': 'str', 'doc': 'save_displayed_chan_only from header global configuration'},
            {'name': 'headstage_firmware_version', 'type': 'str', 'doc': 'headstage_firmware_version from header global configuration'},
            {'name': 'qt_version', 'type': 'str', 'doc': 'qt_version from header global configuration'},
            {'name': 'compile_date', 'type': 'str', 'doc': 'compile_date from header global configuration'},
            {'name': 'compile_time', 'type': 'str', 'doc': 'compile_time from header global configuration'},
            {'name': 'file_prefix', 'type': 'str', 'doc': 'file_prefix from header global configuration'},
            {'name': 'headstage_gyro_sensor_on', 'type': 'str', 'doc': 'headstage_gyro_sensor_on from header global configuration'},
            {'name': 'headstage_mag_sensor_on', 'type': 'str', 'doc': 'headstage_mag_sensor_on from header global configuration'},
            {'name': 'trodes_version', 'type': 'str', 'doc': 'trodes_version from header global configuration'},
            {'name': 'headstage_accel_sensor_on', 'type': 'str', 'doc': 'headstage_accel_sensor_on from header global configuration'},
            {'name': 'commit_head', 'type': 'str', 'doc': 'commit_head from header global configuration'},
            {'name': 'system_time_at_creation', 'type': 'str', 'doc': 'system_time_at_creation from header global configuration'},
            {'name': 'file_path', 'type': 'str', 'doc': 'file_path from header global configuration'}
            ))
    def __init__(self, **kwargs):
        super().__init__(**{kwargs_item: kwargs[kwargs_item]
                            for kwargs_item in kwargs.copy()
                            if kwargs_item not in
                                    ['headstage_serial', 'headstage_smart_ref_on', 'realtime_mode', 'headstage_auto_settle_on',
                                     'timestamp_at_creation', 'controller_firmware_version', 'controller_serial',
                                     'save_displayed_chan_only', 'headstage_firmware_version', 'qt_version', 'compile_date',
                                     'compile_time', 'file_prefix', 'headstage_gyro_sensor_on', 'headstage_mag_sensor_on',
                                     'trodes_version', 'headstage_accel_sensor_on', 'commit_head', 'system_time_at_creation',
                                     'file_path']
                            })
        call_docval_func(super(HeaderDevice, self).__init__, kwargs)
        self.headstage_serial = kwargs['headstage_serial']
        self.headstage_smart_ref_on = kwargs['headstage_smart_ref_on']
        self.realtime_mode = kwargs['realtime_mode']
        self.headstage_auto_settle_on = kwargs['headstage_auto_settle_on']
        self.timestamp_at_creation = kwargs['timestamp_at_creation']
        self.controller_firmware_version = kwargs['controller_firmware_version']
        self.controller_serial = kwargs['controller_serial']
        self.save_displayed_chan_only = kwargs['save_displayed_chan_only']
        self.headstage_firmware_version = kwargs['headstage_firmware_version']
        self.qt_version = kwargs['qt_version']
        self.compile_date = kwargs['compile_date']
        self.compile_time = kwargs['compile_time']
        self.file_prefix = kwargs['file_prefix']
        self.headstage_gyro_sensor_on = kwargs['headstage_gyro_sensor_on']
        self.headstage_mag_sensor_on = kwargs['headstage_mag_sensor_on']
        self.trodes_version = kwargs['trodes_version']
        self.headstage_accel_sensor_on = kwargs['headstage_accel_sensor_on']
        self.commit_head = kwargs['commit_head']
        self.system_time_at_creation = kwargs['system_time_at_creation']
        self.file_path = kwargs['file_path']



# import base64
# import io

# import os
import hither2 as hi
import kachery_client as kc
# import time
from sortingview.helpers.prepare_snippets_h5 import prepare_snippets_h5
import numpy as np
from sortingview.config import job_cache, job_handler
from sortingview.extractors import LabboxEphysRecordingExtractor
from sortingview.serialize_wrapper import serialize_wrapper


@kc.taskfunction('get_timeseries_segment.1', type='pure-calculation')
def task_get_timeseries_segment(recording_object, ds_factor, segment_num, segment_size):
    with hi.Config(job_handler=job_handler.timeseries, job_cache=job_cache):
        return get_timeseries_segment.run(recording_object=recording_object, ds_factor=ds_factor, segment_num=segment_num, segment_size=segment_size)


@hi.function(
    'get_timeseries_segment', '0.1.2',
    image=hi.RemoteDockerImage('docker://magland/labbox-ephys-processing:0.3.19'),
    modules=['sortingview']
)
@serialize_wrapper
def get_timeseries_segment(recording_object, ds_factor, segment_num, segment_size):
    recording0 = LabboxEphysRecordingExtractor(recording_object, download=False)

    t1 = segment_num * segment_size * ds_factor
    t2 = ((segment_num + 1) * segment_size * ds_factor)
    if t2 > recording0.get_num_frames():
        t2 = int(recording0.get_num_frames() / ds_factor) * ds_factor
    traces = recording0.get_traces(
        start_frame=t1,
        end_frame=t2
    )
    M = traces.shape[0]
    N = traces.shape[1]
    if ds_factor > 1:
        N2 = int(N / ds_factor)
        traces_reshaped = traces.reshape((M, N2, ds_factor))
        traces_min = np.min(traces_reshaped, axis=2)
        traces_max = np.max(traces_reshaped, axis=2)
        traces = np.zeros((M, N2 * 2), dtype=np.float32)
        traces[:, 0::2] = traces_min
        traces[:, 1::2] = traces_max
    
    return {
        'traces': traces.astype(np.float32)
    }
    # data_b64 = _mda32_to_base64(traces)
    # # elapsed = time.time() - timer
    # return dict(
    #     data_b64=data_b64
    # )

# def _mda32_to_base64(X) -> str:
#     f = io.BytesIO()
#     le.writemda32(X, f)
#     return base64.b64encode(f.getvalue()).decode('utf-8')



import numpy as np
import kachery_client as kc
import spikeextractors as se

def _load_geom_from_csv(path: str) -> list:
    return _listify_ndarray(np.genfromtxt(path, delimiter=',').T)

def _listify_ndarray(x: np.ndarray) -> list:
    if x.ndim == 1:
        if np.issubdtype(x.dtype, np.integer):
            return [int(val) for val in x]
        else:
            return [float(val) for val in x]
    elif x.ndim == 2:
        ret = []
        for j in range(x.shape[1]):
            ret.append(_listify_ndarray(x[:, j]))
        return ret
    elif x.ndim == 3:
        ret = []
        for j in range(x.shape[2]):
            ret.append(_listify_ndarray(x[:, :, j]))
        return ret
    elif x.ndim == 4:
        ret = []
        for j in range(x.shape[3]):
            ret.append(_listify_ndarray(x[:, :, :, j]))
        return ret
    else:
        raise Exception('Cannot listify ndarray with {} dims.'.format(x.ndim))


def spikeinterface_recording_dict_to_labbox_dict(x):
  c = x['class']
  if c == 'spiketoolkit.preprocessing.bandpass_filter.BandpassFilterRecording':
    kwargs = x['kwargs']
    recording = spikeinterface_recording_dict_to_labbox_dict(kwargs['recording'])
    freq_min = kwargs['freq_min']
    freq_max = kwargs['freq_max']
    freq_wid = kwargs['freq_wid']
    return _make_json_safe({
        'recording_format': 'filtered',
        'data': {
          'filters': [{'type': 'bandpass_filter', 'freq_min': freq_min, 'freq_max': freq_max, 'freq_wid': freq_wid}],
          'recording': recording
        }
    })
  elif c == 'spikeextractors.subrecordingextractor.SubRecordingExtractor':
    kwargs = x['kwargs']
    recording = spikeinterface_recording_dict_to_labbox_dict(kwargs['parent_recording'])
    channel_ids = kwargs['channel_ids']
    renamed_channel_ids = kwargs.get('renamed_channel_ids', None)
    start_frame = kwargs['start_frame']
    end_frame = kwargs['end_frame']
    if renamed_channel_ids is not None:
      raise Exception('renamed_channel_ids field not supported')
    return _make_json_safe({
        'recording_format': 'subrecording',
        'data': {
          'recording': recording,
          'channel_ids': channel_ids,
          'start_frame': start_frame,
          'end_frame': end_frame
        }
    })
  elif c == 'spikeextractors.extractors.mdaextractors.mdaextractors.MdaRecordingExtractor':
    kwargs = x['kwargs']
    path = kwargs['folder_path']
    raw_path = kc.load_file(path + '/raw.mda')
    params_path = path + '/params.json'
    geom_path = path + '/geom.csv'
    params = kc.load_json(params_path)
    assert params is not None, f'Unable to load params.json from: {params_path}'
    geom = _load_geom_from_csv(geom_path)
    return _make_json_safe({
        'recording_format': 'mda',
        'data': {
            'raw': raw_path,
            'geom': geom,
            'params': params
        }
    })
  else:
    raise Exception(f'Unsupported class: {c}')

def _make_json_safe(x):
    if isinstance(x, np.integer):
        return int(x)
    elif isinstance(x, np.floating):
        return float(x)
    elif type(x) == dict:
        ret = dict()
        for key, val in x.items():
            ret[key] = _make_json_safe(val)
        return ret
    elif (type(x) == list) or (type(x) == tuple):
        return [_make_json_safe(val) for val in x]
    elif isinstance(x, np.ndarray):
        raise Exception('Cannot make ndarray json safe')
    else:
        if _is_jsonable(x):
            # this will capture int, float, str, bool
            return x
    raise Exception(f'Item is not json safe: {type(x)}')

def _is_jsonable(x) -> bool:
    import json
    try:
        json.dumps(x)
        return True
    except:
        return False


import hither2 as hi
from sortingview.config import job_cache, job_handler
from sortingview.extractors import LabboxEphysRecordingExtractor, LabboxEphysSortingExtractor
import kachery_client as kc

@hi.function(
    'get_isi_violation_rates', '0.1.1',
    image=hi.RemoteDockerImage('docker://magland/labbox-ephys-processing:0.3.19'),
    modules=['sortingview']
)
def get_isi_violation_rates(sorting_object, recording_object, configuration={}, snippet_len=(50, 80)):
    import spikemetrics as sm
    S = LabboxEphysSortingExtractor(sorting_object)
    R = LabboxEphysRecordingExtractor(recording_object)

    samplerate = R.get_sampling_frequency()
#    duration_sec = R.get_num_frames() / samplerate

    isi_threshold_msec = configuration.get('isi_threshold_msec', 2.5)
    unit_ids = configuration.get('unit_ids', S.get_unit_ids())

    ret = {}
    for id in unit_ids:
        spike_train = S.get_unit_spike_train(unit_id=id)
        ret[str(id)], _ = sm.metrics.isi_violations( #_ is total violations
            spike_train=spike_train,
            duration=R.get_num_frames(),
            isi_threshold=isi_threshold_msec / 1000 * samplerate
        )
    return ret

@kc.taskfunction('get_isi_violation_rates.1', type='pure-calculation')
def task_get_isi_violation_rates(sorting_object, recording_object, configuration={}, snippet_len=(50, 80)):
    with hi.Config(
        job_cache=job_cache,
        job_handler=job_handler.metrics
    ):
        return get_isi_violation_rates.run(
            sorting_object=sorting_object,
            recording_object=recording_object,
            configuration=configuration
        )


from numpy.core.records import record
import spikeextractors as se
import numpy as np
import hither2 as hi
import kachery_client as kc
from sortingview.config import job_cache, job_handler
from sortingview.extractors import LabboxEphysRecordingExtractor, LabboxEphysSortingExtractor

# adjust these values
workspace_uri = '{workspaceUri}'
recording_label = 'simulated_recording'
duration_sec = 60 # duration of simulated recording
num_channels = 8 # num. channels in simulated recording
num_units = 5 # num units
seed = 1 # random number generator seed

def prepare_recording_sorting():
    # Simulate a recording (toy example)
    recording, sorting = se.example_datasets.toy_example(duration=duration_sec, num_channels=num_channels, K=num_units, seed=seed)
    R = LabboxEphysRecordingExtractor.from_memory(recording, serialize=True, serialize_dtype=np.int16)
    S = LabboxEphysSortingExtractor.from_memory(sorting, serialize=True)
    return R, S


@hi.function('example_recording_sortings', '0.1.2')
def example_recording_sortings():
    recording, sorting_true = prepare_recording_sorting()
    recording_uri = kc.store_json(recording.object(), basename='example_recording.json')
    sorting_uri = kc.store_json(sorting_true.object(), basename='example_sorting_true.json')
    return [{
        'label': 'Simulation1',
        'recordingObject': recording.object(),
        'recordingUri': recording_uri,
        'sortingObject': sorting_true.object(),
        'sortingUri': sorting_uri
    }]


# @kc.taskfunction('example_recording_sortings', type='pure-calculation')
def task_example_recording_sortings(cachebust: str):
    with hi.Config(job_handler=job_handler.misc, job_cache=job_cache):
        return hi.Job(example_recording_sortings, {})


from typing import Dict, List, Union

import math
import numpy as np
import h5py

def _intersect_channel_ids(channel_ids_list: List[List[int]]):
    # from js:
    #   channel_ids_list is a list of the channel IDs that make up the neighborhood for each unit. This function will be passed a set of unit IDs that are to be merged, along with their respective channel neighborhoods. Its job is to reduce that list to only those channels in the neighborhood of each of the underlying units.
    #   It is assumed that there will be at least one channel common to all units being merged (so that x in line 10 never collapses to an empty list); violating this assumption leads to failure of the function.
    #   inds will actually wind up being a list of lists--the per-unit-neighborhood indices of the channels in the consolidated/fully intersected set of channels. Like the input channel_ids_list, its outermost index will be the units, so the channel indices can always be mapped back to the units whose neighborhoods they're identifiying.
    
    x = channel_ids_list[0]
    for y in channel_ids_list:
        x = np.intersect1d(x, y).tolist()
    inds = []
    for y in channel_ids_list:
        a, inds_x, inds_y = np.intersect1d(x, y, return_indices=True)
        inds.append(inds_y.tolist())
    return x, inds

def _get_unit_waveforms_from_list_from_snippets_h5(snippets_h5_path: str, unit_ids: List[int], max_num_events: int = None):
    unit_waveforms_list = []
    unit_waveforms_channel_ids_list = []
    channel_locations_list = []
    sampling_frequency_list = []
    unit_spike_train_list = []
    for unit_id in unit_ids:
        unit_waveforms0, unit_waveforms_channel_ids0, channel_locations0, sampling_frequency0, unit_spike_train0 = get_unit_waveforms_from_snippets_h5(snippets_h5_path, unit_id, max_num_events=max_num_events)
        unit_waveforms_list.append(unit_waveforms0)
        unit_waveforms_channel_ids_list.append(unit_waveforms_channel_ids0)
        channel_locations_list.append(channel_locations0)
        sampling_frequency_list.append(sampling_frequency0)
        unit_spike_train_list.append(unit_spike_train0)
    intersection_channel_ids, inds = _intersect_channel_ids(unit_waveforms_channel_ids_list)
    for i in range(len(unit_ids)):
        unit_waveforms_list[i] = unit_waveforms_list[i][:, inds[i]]
        unit_waveforms_channel_ids_list[i] = intersection_channel_ids
        channel_locations_list[i] = channel_locations_list[i][inds[i]]
    unit_waveforms = np.concatenate(unit_waveforms_list, axis=0)
    unit_spike_train = np.concatenate(unit_spike_train_list)
    return unit_waveforms, np.array(intersection_channel_ids), channel_locations_list[0], sampling_frequency_list[0], unit_spike_train

def subsample_inds(n, m):
    if m >= n:
        return range(n)
    incr = n / m
    return [int(math.floor(i * incr)) for i in range(m)]

def get_unit_waveforms_from_snippets_h5(snippets_h5_path: str, unit_id: Union[int, List[int]], max_num_events: int = None):
    if type(unit_id) is list:
        return _get_unit_waveforms_from_list_from_snippets_h5(snippets_h5_path, unit_id, max_num_events=max_num_events)
    with h5py.File(snippets_h5_path, 'r') as f:
        channel_ids = np.array(f.get('channel_ids'))
        channel_locations = np.array(f.get(f'channel_locations'))
        unit_spike_train = np.array(f.get(f'unit_spike_trains/{unit_id}'))
        sampling_frequency = np.array(f.get('sampling_frequency'))[0].item()
        if np.isnan(sampling_frequency):
            print('WARNING: sampling frequency is nan. Using 30000 for now. Please correct the snippets file.')
            sampling_frequency = 30000
        unit_waveforms = np.array(f.get(f'unit_waveforms/{unit_id}/waveforms'))
        unit_waveforms_channel_ids = np.array(f.get(f'unit_waveforms/{unit_id}/channel_ids'))

        if max_num_events is not None:
            if len(unit_spike_train) > max_num_events:
                inds = subsample_inds(len(unit_spike_train), max_num_events)
                unit_spike_train = unit_spike_train[inds]
                unit_waveforms = unit_waveforms[inds]

    channel_locations0 = []
    for ch_id in unit_waveforms_channel_ids:
        ind = np.where(channel_ids == ch_id)[0]
        channel_locations0.append(channel_locations[ind, :].ravel().tolist())
    
    return unit_waveforms, unit_waveforms_channel_ids, np.array(channel_locations0), sampling_frequency, unit_spike_train



from typing import List, Set, Union

import h5py
import kachery_client as kc
import numpy as np
import spikeextractors as se


class Snippets1SortingExtractor(se.SortingExtractor):
    extractor_name = 'Snippets1SortingExtractor'
    is_writable = False
    def __init__(self, *, snippets_h5_uri: str, p2p: bool=False):
        se.RecordingExtractor.__init__(self)

        snippets_h5_path = kc.load_file(snippets_h5_uri, p2p=p2p)
        
        self._snippets_h5_path: str = snippets_h5_path

        channel_ids_set: Set[int] = set()
        max_timepoint: int = 0
        with h5py.File(self._snippets_h5_path, 'r') as f:
            sampling_frequency: float = np.array(f.get('sampling_frequency'))[0]
            if np.isnan(sampling_frequency):
                print('WARNING: sampling frequency is nan. Using 30000 for now. Please correct the snippets file.')
                sampling_frequency = 30000
            self.set_sampling_frequency(sampling_frequency)
            self._unit_ids: List[int] = np.array(f.get('unit_ids')).astype(int).tolist()
            for unit_id in self._unit_ids:
                unit_spike_train = np.array(f.get(f'unit_spike_trains/{unit_id}'))
                max_timepoint = int(max(max_timepoint, np.max(unit_spike_train)))
                # unit_waveforms = np.array(f.get(f'unit_waveforms/{unit_id}/waveforms'))
                unit_waveforms_channel_ids = np.array(f.get(f'unit_waveforms/{unit_id}/channel_ids'))
                for id in unit_waveforms_channel_ids:
                    channel_ids_set.add(int(id))
        self._channel_ids: List[int] = sorted(list(channel_ids_set))
        self._num_frames: int = max_timepoint + 1

    def get_unit_ids(self) -> List[int]:
        return self._unit_ids

    def get_unit_spike_train(self, unit_id: int, start_frame: Union[None, int], end_frame: Union[None, int]) -> int:
        with h5py.File(self._snippets_h5_path, 'r') as f:
            unit_spike_train = np.array(f.get(f'unit_spike_trains/{unit_id}'))
            if start_frame is not None:
                assert end_frame is not None
                return unit_spike_train[(start_frame <= unit_spike_train) & (unit_spike_train < end_frame)]
            else:
                return unit_spike_train

    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None, return_scaled=True):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = self._num_frames
        if channel_ids is None:
            channel_ids = self._channel_ids
        M = len(channel_ids)
        N = end_frame - start_frame

        # For now, just return zeros
        return np.zeros((M, N))



from typing import List, cast
import hither2 as hi
import numpy as np
import spikeextractors as se
import kachery_client as kc
from sortingview.config import job_cache, job_handler
from sortingview.extractors import LabboxEphysSortingExtractor

@hi.function(
    'sorting_info', '0.1.3'
)
def sorting_info(sorting_uri):
    sorting = LabboxEphysSortingExtractor(sorting_uri)
    return dict(
        unit_ids=_to_int_list(sorting.get_unit_ids()),
        samplerate=sorting.get_sampling_frequency(),
        sorting_object=sorting.object()
    )

@kc.taskfunction('sorting_info.3', type='pure-calculation')
def task_sorting_info(sorting_uri: str):
    with hi.Config(job_handler=job_handler.misc, job_cache=job_cache):
        return hi.Job(sorting_info, {'sorting_uri': sorting_uri})


def _to_int_list(x):
    return np.array(x).astype(int).tolist()


from typing import Dict, Union

import os
import hither2 as hi
import kachery_client as kc
import numpy as np
import spikeextractors as se
from sortingview.extractors import LabboxEphysSortingExtractor, LabboxEphysRecordingExtractor
from .SubsampledSortingExtractor import SubsampledSortingExtractor
from .find_unit_peak_channels import find_unit_peak_channels
from .find_unit_neighborhoods import find_unit_neighborhoods
from .get_unit_waveforms import get_unit_waveforms

@hi.function(
    'prepare_snippets_h5', '0.2.7',
    image=hi.RemoteDockerImage('docker://magland/labbox-ephys-processing:0.3.19'),
    modules=['sortingview']
)
def prepare_snippets_h5(
    recording_object,
    sorting_object,
    start_frame=None,
    end_frame=None,
    max_events_per_unit=None,
    max_neighborhood_size=15,
    snippet_len=(50, 80)
):
    if recording_object['recording_format'] == 'snippets1':
        return recording_object['data']['snippets_h5_uri']

    recording = LabboxEphysRecordingExtractor(recording_object)
    sorting = LabboxEphysSortingExtractor(sorting_object)

    with kc.TemporaryDirectory() as tmpdir:
        save_path = tmpdir + '/snippets.h5'
        prepare_snippets_h5_from_extractors(
            recording=recording,
            sorting=sorting,
            output_h5_path=save_path,
            start_frame=start_frame,
            end_frame=end_frame,
            max_events_per_unit=max_events_per_unit,
            max_neighborhood_size=max_neighborhood_size,
            snippet_len=snippet_len
        )
        return kc.store_file(save_path)

def prepare_snippets_h5_from_extractors(
    recording: se.RecordingExtractor,
    sorting: se.SortingExtractor,
    output_h5_path: str,
    start_frame,
    end_frame,
    max_neighborhood_size: int,
    max_events_per_unit: Union[None, int]=None,
    snippet_len=(50, 80)
):
    import h5py
    if start_frame is not None:
        recording = se.SubRecordingExtractor(parent_recording=recording, start_frame=start_frame, end_frame=end_frame)
        sorting = se.SubSortingExtractor(parent_sorting=sorting, start_frame=start_frame, end_frame=end_frame)

    unit_ids = sorting.get_unit_ids()
    samplerate = recording.get_sampling_frequency()
    
    # Use this optimized function rather than spiketoolkit's version
    # for efficiency with long recordings and/or many channels, units or spikes
    # we should submit this to the spiketoolkit project as a PR
    print('Subsampling sorting')
    if max_events_per_unit is not None:
        sorting_subsampled = SubsampledSortingExtractor(parent_sorting=sorting, max_events_per_unit=max_events_per_unit, method='random')
    else:
        sorting_subsampled = sorting
    print('Finding unit peak channels')
    peak_channels_by_unit = find_unit_peak_channels(recording=recording, sorting=sorting, unit_ids=unit_ids)
    print('Finding unit neighborhoods')
    channel_ids_by_unit = find_unit_neighborhoods(recording=recording, peak_channels_by_unit=peak_channels_by_unit, max_neighborhood_size=max_neighborhood_size)
    print(f'Getting unit waveforms for {len(unit_ids)} units')
    unit_waveforms = get_unit_waveforms(
        recording=recording,
        sorting=sorting_subsampled,
        unit_ids=unit_ids,
        channel_ids_by_unit=channel_ids_by_unit,
        snippet_len=snippet_len
    )
    # unit_waveforms = st.postprocessing.get_unit_waveforms(
    #     recording=recording,
    #     sorting=sorting,
    #     unit_ids=unit_ids,
    #     ms_before=1,
    #     ms_after=1.5,
    #     max_spikes_per_unit=500
    # )

    save_path = output_h5_path
    with h5py.File(save_path, 'w') as f:
        f.create_dataset('unit_ids', data=np.array(unit_ids).astype(np.int32))
        f.create_dataset('sampling_frequency', data=np.array([samplerate]).astype(np.float64))
        f.create_dataset('channel_ids', data=np.array(recording.get_channel_ids()))
        f.create_dataset('num_frames', data=np.array([recording.get_num_frames()]).astype(np.int32))
        channel_locations = recording.get_channel_locations()
        f.create_dataset(f'channel_locations', data=np.array(channel_locations))
        for ii, unit_id in enumerate(unit_ids):
            x = sorting.get_unit_spike_train(unit_id=unit_id)
            f.create_dataset(f'unit_spike_trains/{unit_id}', data=np.array(x).astype(np.float64))
            f.create_dataset(f'unit_waveforms/{unit_id}/waveforms', data=unit_waveforms[ii].astype(np.float32))
            f.create_dataset(f'unit_waveforms/{unit_id}/channel_ids', data=np.array(channel_ids_by_unit[int(unit_id)]).astype(int))
            f.create_dataset(f'unit_waveforms/{unit_id}/spike_train', data=np.array(sorting_subsampled.get_unit_spike_train(unit_id=unit_id)).astype(np.float64))



from setuptools import setup, find_packages

setup(
    packages=find_packages(),
    scripts=[
        'bin/sortingview-start-backend'
    ],
    include_package_data = True,
    install_requires=[
        'click',
        'kachery-client>=1.0.19',
        'figurl>=0.1.4',
        'hither>=0.8.1',
        'pynwb',
        'pyyaml',
        'spikeextractors>=0.9.6',
        'spikecomparison',
        'seriesview>=0.1.0',
        'spikeinterface>=0.90.1'
    ]
)



from typing import List, Union
from .labboxephysrecordingextractor import LabboxEphysRecordingExtractor

def subrecording(*, recording: LabboxEphysRecordingExtractor, channel_ids: Union[List[int], None]=None, start_frame: Union[int, None]=None, end_frame: Union[int, None]=None):
    data = {
        'recording': recording.object(),
    }
    if channel_ids is not None:
        data['channel_ids'] = channel_ids
    if start_frame is not None:
        data['start_frame'] = start_frame
        assert end_frame is not None
    if end_frame is not None:
        data['end_frame'] = end_frame
        assert start_frame is not None
    return LabboxEphysRecordingExtractor({
        'recording_format': 'subrecording',
        'data': data
    })


from typing import List, Union
from .labboxephyssortingextractor import LabboxEphysSortingExtractor

def subsorting(*, sorting: LabboxEphysSortingExtractor, unit_ids: Union[List[int], None]=None, start_frame: Union[int, None]=None, end_frame: Union[int, None]=None):
    data = {
        'sorting': sorting.object(),
    }
    if unit_ids is not None:
        data['unit_ids'] = unit_ids
    if start_frame is not None:
        data['start_frame'] = start_frame
        assert end_frame is not None
    if end_frame is not None:
        data['end_frame'] = end_frame
        assert start_frame is not None
    return LabboxEphysSortingExtractor({
        'sorting_format': 'subsorting',
        'data': data
    })


import kachery_client as kc
from ..workspace import load_workspace

@kc.taskfunction('sortingview_workspace_sorting_curation_action.1', type='action')
def task_sortingview_workspace_sorting_curation_action(workspace_uri: str, sorting_id: str, action: dict):
    W = load_workspace(workspace_uri)
    f = W.feed
    curation_subfeed = f.load_subfeed({'name': 'sortingCuration', 'sortingId': sorting_id})
    curation_subfeed.append_message(action)


import numpy as np
import spikeextractors as se

def create_recording_from_old_extractor(rx: se.RecordingExtractor):
    import spikeinterface as si
    class RecordingSegmentWrapper(si.BaseRecordingSegment):
        def __init__(self, rx: se.RecordingExtractor):
            si.BaseRecordingSegment.__init__(self)
            self._rx = rx
            self._channel_ids = np.array(rx.get_channel_ids())

        def get_num_samples(self):
            return self._rx.get_num_frames()

        def get_traces(self, start_frame, end_frame, channel_indices):
            if channel_indices is None:
                channel_ids = self._channel_ids
            else:
                channel_ids = self._channel_ids[channel_indices]
            return self._rx.get_traces(
                channel_ids=channel_ids,
                start_frame=start_frame,
                end_frame=end_frame
            ).T
            
    R = si.BaseRecording(
        sampling_frequency=rx.get_sampling_frequency(),
        channel_ids=rx.get_channel_ids(),
        dtype=np.int16
    )
    recording_segment = RecordingSegmentWrapper(rx)
    R.add_recording_segment(recording_segment)
    return R


import kachery_client as kc
from ..workspace_list import WorkspaceList

# @kc.taskfunction('sortingview_workspace_list_subfeed.2', type='query')
def task_sortingview_workspace_list_subfeed(name: str):
    W = WorkspaceList(list_name=name)
    return W.get_subfeed_uri()


from typing import Dict, List, Union

import hither2 as hi
from hither2.dockerimage import RemoteDockerImage
import kachery_client as kc
import numpy as np
from sortingview.serialize_wrapper import serialize_wrapper
# from labbox import LabboxContext
from sortingview.config import job_cache, job_handler
from sortingview.helpers import prepare_snippets_h5, get_unit_waveforms_from_snippets_h5

@hi.function(
    'fetch_average_waveform_2', '0.2.14',
    image=RemoteDockerImage('docker://magland/labbox-ephys-processing:0.3.19'),
    modules=['sortingview']
)
@serialize_wrapper
def fetch_average_waveform_2(snippets_h5, unit_id):
    import h5py
    h5_path = kc.load_file(snippets_h5)
    assert h5_path is not None
    unit_waveforms, unit_waveforms_channel_ids, channel_locations0, sampling_frequency, unit_spike_train = get_unit_waveforms_from_snippets_h5(h5_path, unit_id)
    
    average_waveform = np.mean(unit_waveforms, axis=0)

    return dict(
        average_waveform=average_waveform.astype(np.float32),
        channel_ids=unit_waveforms_channel_ids.astype(np.int32),
        channel_locations=channel_locations0.astype(np.float32),
        sampling_frequency=sampling_frequency
    )

@kc.taskfunction('fetch_average_waveform.2', type='pure-calculation')
def task_fetch_average_waveform(recording_object, sorting_object, unit_id, snippet_len=(50, 80)):
    with hi.Config(job_handler=job_handler.waveforms, job_cache=job_cache):
        with hi.Config(job_handler=job_handler.extract_snippets):
            snippets_h5 = prepare_snippets_h5.run(recording_object=recording_object, sorting_object=sorting_object, snippet_len=snippet_len)
        return fetch_average_waveform_2.run(
            snippets_h5=snippets_h5,
            unit_id=unit_id
        )

@hi.function('test_delay', '0.1.0')
def test_delay(delay_sec):
    import time
    time.sleep(delay_sec)
    return f'delayed: {delay_sec} sec'

@kc.taskfunction('test_delay.1', type='pure-calculation')
def task_test_delay(delay_sec, cachebust):
    with hi.Config(job_handler=job_handler.waveforms, job_cache=None):
        return test_delay.run(
            delay_sec=delay_sec
        )


from hdmf import docval
from hdmf.utils import get_docval, call_docval_func
from pynwb import register_class
from pynwb.ecephys import ElectrodeGroup


@register_class('NwbElectrodeGroup', 'ndx-franklab-novela')
class NwbElectrodeGroup(ElectrodeGroup):
    ''' Representation of custom ElectrodeGroup object in NWB '''

    __nwbfields__ = ('targeted_location', 'targeted_x', 'targeted_y', 'targeted_z', 'units')

    @docval(*get_docval(ElectrodeGroup.__init__) + (
            {'name': 'targeted_location', 'type': 'str', 'doc': 'predicted location'},
            {'name': 'targeted_x', 'type': 'float', 'doc': 'predicted x coordinates'},
            {'name': 'targeted_y', 'type': 'float', 'doc': 'predicted y coordinates'},
            {'name': 'targeted_z', 'type': 'float', 'doc': 'predicted z coordinates'},
            {'name': 'units', 'type': 'str', 'doc': 'units of fields, possible value: um or mm'}))
    def __init__(self, **kwargs):
        super().__init__(**{kwargs_item: kwargs[kwargs_item]
                            for kwargs_item in kwargs.copy()
                            if kwargs_item != 'targeted_location'
                            if kwargs_item != 'targeted_x'
                            if kwargs_item != 'targeted_y'
                            if kwargs_item != 'targeted_z'
                            if kwargs_item != 'units'
                            })
        call_docval_func(super(NwbElectrodeGroup, self).__init__, kwargs)
        self.targeted_location = kwargs['targeted_location']
        self.targeted_x = kwargs['targeted_x']
        self.targeted_y = kwargs['targeted_y']
        self.targeted_z = kwargs['targeted_z']
        self.units = kwargs['units']



import os
from typing import Union
import kachery_client as kc
import yaml

from ..workspace_list import get_workspace_list, set_workspace_list
from ._verify_oauth2_token import _verify_oauth2_token

auth_config_path = os.getenv('SORTINGVIEW_AUTH_CONFIG', None)
if auth_config_path is not None:
    print(f'Using auth config file: {auth_config_path}')
    with open(auth_config_path, 'r') as f:
        auth_config = yaml.safe_load(f)
else:
    print('Using default auth config. To override, set SORTINGVIEW_AUTH_CONFIG to path of a yaml file.')
    auth_config = {}

workspace_lists_auth_config = auth_config.get('workspaceLists', None)

# @kc.taskfunction('sortingview.get_workspace_list.1', type='query')
def task_get_workspace_list(name: str, id_token: Union[str, None]=None):
    if id_token is not None:
        id_info = _verify_oauth2_token(id_token.encode('utf-8'))
        auth_user_id = id_info['email']
    else:
        auth_user_id = None
    if workspace_lists_auth_config is not None:
        if auth_user_id is None: raise Exception('Not authorized (not logged in)')
        ok = False
        for list_name, c in workspace_lists_auth_config.items():
            if list_name == name:
                for x in c:
                    if x['userId'] == auth_user_id:
                        ok = True
    else:
        ok = True
    if not ok: raise Exception('Not authorized')
    workspace_list = get_workspace_list(name=name)
    return workspace_list


# This file was automatically generated by jinjaroot. Do not edit directly.
import click
from .start_backend import start_backend

@click.command('sortingview-start-backend')
@click.option('--channel', required=True, help="The kachery channel")
@click.option('--backend-id', required=False, default='', help="The backend ID")
def start_backend_cli(channel: str, backend_id: str):
    bid = backend_id if backend_id != '' else None
    start_backend(channel=channel, backend_id=bid)